# -*- encoding: utf-8 -*-

from __future__ import unicode_literals

import re
import pytest

from hyputils.memex.util import document_claims
from hyputils.memex.util.document_claims import doi_uri_from_string


class TestDocumentURIsFromLinks(object):
    def test_it_ignores_href_links_that_match_the_claimant_uri(self):
        """
        Links containing only the claimant URI should be ignored.

        If document.link contains a link dict with just an "href" and no other
        keys, and the value of the "href" key is the same as the claimant URI,
        then this link dict should be ignored and not produce an additional
        document URI dict in the output (since the document URI that it would
        generate would be the same as the "self-claim" claimant URI one that is
        always generated anyway).

        """
        claimant = "http://localhost:5000/docs/help"
        link_dicts = [{"href": claimant}]

        document_uris = document_claims.document_uris_from_links(link_dicts, claimant)

        assert document_uris == []

    def test_it_ignores_doi_links(self):
        """
        Links containing only an href that starts with doi should be ignored.

        If document.link contains a link dict with just an "href" and no other
        keys, and the value of the "href" key begins with "doi:", then the link
        dict should be ignored and not produce a document URI dict in the
        output.

        This is because document URI dicts for doi: URIs are generate
        separately from other metadata in the document dict outside of the
        "link" list.

        """
        link_dicts = [{"href": "doi:10.3389/fenvs.2014.00003"}]

        document_uris = document_claims.document_uris_from_links(
            link_dicts, claimant="http://localhost:5000/docs/help"
        )

        assert document_uris == []

    def test_it_ignores_highwire_pdf_links(self):
        pdf_url = "http://example.com/example.pdf"
        link_dicts = [{"href": pdf_url, "type": "application/pdf"}]

        document_uris = document_claims.document_uris_from_links(
            link_dicts, claimant="http://localhost:5000/docs/help"
        )

        assert document_uris == []

    def test_it_returns_rel_alternate_document_uris_for_rel_alternate_links(self):
        alternate_url = "http://example.com/alternate"
        link_dicts = [{"href": alternate_url, "rel": "alternate"}]

        document_uris = document_claims.document_uris_from_links(
            link_dicts, claimant="http://localhost:5000/docs/help"
        )

        alternate_document_uri = one(
            [d for d in document_uris if d["type"] == "rel-alternate"]
        )
        assert alternate_document_uri == {
            "type": "rel-alternate",
            "claimant": "http://localhost:5000/docs/help",
            "content_type": "",
            "uri": alternate_url,
        }

    def test_it_uses_link_types_as_document_uri_content_types(self):
        """
        Link types get converted to document URI content_types.

        The value of the 'type' key in link dicts ends up as the value of the
        'content_type' key in the returned document URI dicts.

        """
        link_dicts = [{"href": "http://example.com/example.html", "type": "text/html"}]

        document_uris = document_claims.document_uris_from_links(
            link_dicts, claimant="http://example.com/example.html"
        )

        assert one([d for d in document_uris if d.get("content_type") == "text/html"])

    def test_it_returns_multiple_document_URI_dicts(self):  # noqa
        """If there are multiple claims it should return multiple dicts."""
        link_dicts = [
            {"href": "http://example.com/example.html", "type": "text/html"},
            {"href": "http://example.com/alternate.html", "rel": "alternate"},
            {"href": "http://example.com/example2.html", "type": "text/html"},
        ]

        document_uris = document_claims.document_uris_from_links(
            link_dicts, claimant="http://example.com/claimant.html"
        )

        assert len(document_uris) == 3


class TestDocumentMetasFromData(object):
    @pytest.mark.parametrize(
        "input_,output",
        [
            # String values get turned into length 1 lists.
            ({"foo": "string"}, {"type": "foo", "value": ["string"]}),
            # List values get copied over unchanged.
            ({"foo": ["one", "two"]}, {"type": "foo", "value": ["one", "two"]}),
            # Sub-dicts get flattened using a '.' separator in the key,
            # and length 1 list values in sub-dicts get copied over unchanged.
            (
                {"facebook": {"description": ["document description"]}},
                {"type": "facebook.description", "value": ["document description"]},
            ),
            # Length >1 list values in sub-dicts get copied over unchanged.
            (
                {
                    "facebook": {
                        "image": [
                            "http://example.com/image1.png",
                            "http://example.com/image2.png",
                            "http://example.com/image3.jpeg",
                        ]
                    }
                },
                {
                    "type": "facebook.image",
                    "value": [
                        "http://example.com/image1.png",
                        "http://example.com/image2.png",
                        "http://example.com/image3.jpeg",
                    ],
                },
            ),
            # String values in sub-dicts get turned into length 1 lists.
            ({"foo": {"bar": "string"}}, {"type": "foo.bar", "value": ["string"]}),
            # Leading and trailing whitespace gets stripped from document titles.
            (
                {
                    "title": [
                        "   My Document",
                        "My Document   ",
                        " My Document ",
                        "\nMy Document\n\n",
                        "\rMy Document\r\n",
                        "\tMy Document \t \t ",
                    ]
                },
                {
                    "type": "title",
                    "value": [
                        "My Document",
                        "My Document",
                        "My Document",
                        "My Document",
                        "My Document",
                        "My Document",
                    ],
                },
            ),
            # Leading and trailing whitespace does not get-stripped from non-titles.
            (
                {
                    "foo": [
                        "   My Document",
                        "My Document   ",
                        " My Document ",
                        "\nMy Document\n\n",
                        "\rMy Document\r\n",
                        "\tMy Document \t \t ",
                    ]
                },
                {
                    "type": "foo",
                    "value": [
                        "   My Document",
                        "My Document   ",
                        " My Document ",
                        "\nMy Document\n\n",
                        "\rMy Document\r\n",
                        "\tMy Document \t \t ",
                    ],
                },
            ),
        ],
    )
    def test_document_metas_from_data(self, input_, output):
        claimant = "http://example.com/claimant/"

        document_metas = document_claims.document_metas_from_data(
            document_data=input_, claimant=claimant
        )

        assert document_metas == [
            {"type": output["type"], "value": output["value"], "claimant": claimant}
        ]

    def test_document_metas_from_data_ignores_links_list(self):
        """It should ignore the "link" list in the document_data."""
        document_data = {"link": [{"href": "http://example.com/link"}]}

        document_metas = document_claims.document_metas_from_data(
            document_data, "http://example/claimant"
        )

        assert document_metas == []

    def test_document_metas_from_data_with_multiple_metadata_claims(self):
        """
        It should create one DocumentMeta for each metadata claim.

        If document_data contains multiple metadata claims it should init one
        DocumentMeta for each claim.

        """
        claimant = "http://example/claimant"
        document_data = {
            "title": "the title",
            "description": "the description",
            "site_title": "the site title",
        }

        document_metas = document_claims.document_metas_from_data(
            document_data, claimant
        )

        assert len(document_metas) == len(document_data.items())
        for key, value in document_data.items():
            assert {
                "type": key,
                "value": [value],
                "claimant": claimant,
            } in document_metas

    def test_document_metas_from_data_ignores_null_titles(self):
        """It should ignore null document titles."""
        for title in (None, [None, None]):
            document_data = {"title": title}

            document_metas = document_claims.document_metas_from_data(
                document_data, "http://example/claimant"
            )

            assert document_metas == []

    def test_document_metas_from_data_allows_null_non_titles(self):
        """Null values are allowed if 'type' isn't 'title'."""
        for value in (None, [None, None]):
            document_data = {"foo": value}

            document_metas = document_claims.document_metas_from_data(
                document_data, "http://example/claimant"
            )

            if not isinstance(value, list):
                # We expect it to turn non-lists into length-1 lists.
                value = [value]

            assert document_metas == [
                {"type": "foo", "value": value, "claimant": "http://example/claimant"}
            ]

    def test_document_metas_from_data_ignores_empty_string_titles(self):
        """It should ignore empty document titles."""
        for title in ("", ["", ""]):
            document_data = {"title": title}

            document_metas = document_claims.document_metas_from_data(
                document_data, "http://example/claimant"
            )

            assert document_metas == []

    def test_document_metas_from_data_allows_empty_string_non_titles(self):
        """Empty strings are allowed if 'type' isn't 'title'."""
        for value in ("", ["", ""]):
            document_data = {"foo": value}

            document_metas = document_claims.document_metas_from_data(
                document_data, "http://example/claimant"
            )

            if not isinstance(value, list):
                # We expect it to turn non-lists into length-1 lists.
                value = [value]

            assert document_metas == [
                {"type": "foo", "value": value, "claimant": "http://example/claimant"}
            ]

    def test_document_metas_from_data_ignores_whitespace_only_titles(self):
        """It should ignore whitespace-only document titles."""
        for title in (" ", [" ", " "], "\n\n  \n"):
            document_data = {"title": title}

            document_metas = document_claims.document_metas_from_data(
                document_data, "http://example/claimant"
            )

            assert document_metas == []

    def test_document_metas_from_data_allows_whitespace_only_non_titles(self):
        """Whitespace-only strings are allowed if 'type' isn't 'title'."""
        for value in (" ", [" ", " "], "\n\n  \n"):
            document_data = {"foo": value}

            document_metas = document_claims.document_metas_from_data(
                document_data, "http://example/claimant"
            )

            if not isinstance(value, list):
                # We expect it to turn non-lists into length-1 lists.
                value = [value]

            assert document_metas == [
                {"type": "foo", "value": value, "claimant": "http://example/claimant"}
            ]


class TestDocumentURIsFromHighwirePDF(object):
    def test_highwire_pdf_values_produce_highwire_pdf_document_uris(self):
        highwire_dict = {
            "pdf_url": [
                "http://example.com/1.pdf",
                "http://example.com/2.pdf",
                "http://example.com/3.pdf",
            ]
        }

        document_uris = document_claims.document_uris_from_highwire_pdf(
            highwire_dict, claimant="http://example.com/example.html"
        )

        for pdf in highwire_dict["pdf_url"]:
            document_uri = one([d for d in document_uris if d.get("uri") == pdf])
            assert document_uri == {
                "claimant": "http://example.com/example.html",
                "uri": pdf,
                "type": "highwire-pdf",
                "content_type": "application/pdf",
            }


class TestDOIURIFromString(object):
    @pytest.mark.parametrize("doi", ["10.1001/1234", "doi:10.1001/1234"])
    def test_it_prepends_doi_prefix(self, doi):
        assert doi_uri_from_string(doi) == "doi:{}".format(strip_prefix("doi:", doi))

    @pytest.mark.parametrize(
        "url",
        [
            "http://doi.org/10.1234/5678",
            "https://doi.org/10.1234/5678",
            "http://dx.doi.org/10.1234/5678",
            "https://dx.doi.org/10.1234/5678",
        ],
    )
    def test_it_allows_doi_urls(self, url):
        # Many sites store DOI URLs rather than just identifiers in DOI fields.
        # We should ideally normalize the different forms, but for now we just
        # continue to accept them.
        assert doi_uri_from_string(url) == "doi:{}".format(url)

    @pytest.mark.parametrize(
        "doi",
        [
            # Empty
            "doi:",
            "",
            # Whitespace only
            "doi: ",
            " ",
            # Strings that do not match the DOI syntax.
            "9.1234/567",
            "chapter1/section1",
            "1234.5678",
            "10.0.0.1",
            "10.0/1234",
            # Non-DOI URLs
            "https://publisher.org/foo.html",
        ],
    )
    def test_it_returns_none_if_invalid(self, doi):
        assert doi_uri_from_string(doi) is None

    @pytest.mark.parametrize("doi", ["  doi: 10.1234/5678"])
    def test_it_strips_whitespace(self, doi):
        assert doi_uri_from_string(doi) == re.sub("\\s+", "", doi)


class TestDocumentURIsFromHighwireDOI(object):
    def test_highwire_doi_values_produce_highwire_doi_document_uris(self):
        highwire_dict = {
            "doi": [
                "doi:10.1038/nphys1170",
                "doi:10.1002/0470841559.ch1",
                "doi:10.1594/PANGAEA.726855",
            ]
        }

        document_uris = document_claims.document_uris_from_highwire_doi(
            highwire_dict, claimant="http://example.com/example.html"
        )

        for doi in highwire_dict["doi"]:
            document_uri = one([d for d in document_uris if d.get("uri") == doi])
            assert document_uri == {
                "claimant": "http://example.com/example.html",
                "uri": doi,
                "type": "highwire-doi",
                "content_type": "",
            }

    def test_it_ignores_invalid_dois(self):
        """If `doi_uri_from_string` returns `None`, the identifier is ignored."""
        highwire_dict = {"doi": ["doi:"]}
        document_uris = document_claims.document_uris_from_highwire_doi(
            highwire_dict, claimant="http://example.com/example.html"
        )
        assert len(document_uris) == 0


class TestDocumentURIsFromDC(object):
    def test_dc_identifiers_produce_dc_doi_document_uris(self):
        """Each 'identifier' list item in the 'dc' dict becomes a doc URI."""
        dc_dict = {
            "identifier": [
                "doi:10.1038/nphys1170",
                "doi:10.1002/0470841559.ch1",
                "doi:10.1594/PANGAEA.726855",
            ]
        }

        document_uris = document_claims.document_uris_from_dc(
            dc_dict, claimant="http://example.com/example.html"
        )

        for doi in dc_dict["identifier"]:
            document_uri = one([d for d in document_uris if d.get("uri") == doi])
            assert document_uri == {
                "claimant": "http://example.com/example.html",
                "uri": doi,
                "type": "dc-doi",
                "content_type": "",
            }

    def test_it_ignores_invalid_dois(self):
        """If `doi_uri_from_string` returns `None`, the identifier is ignored."""
        dc_dict = {"identifier": ["doi:"]}
        document_uris = document_claims.document_uris_from_dc(
            dc_dict, claimant="http://example.com/example.html"
        )
        assert len(document_uris) == 0


class TestDocumentURISelfClaim(object):
    def test_document_uri_self_claim(self):
        claimant = "http://localhost:5000/docs/help"

        document_uri = document_claims.document_uri_self_claim(claimant)

        assert document_uri == {
            "claimant": claimant,
            "uri": claimant,
            "type": "self-claim",
            "content_type": "",
        }


@pytest.mark.usefixtures(
    "document_uris_from_dc",
    "document_uris_from_highwire_doi",
    "document_uris_from_highwire_pdf",
    "document_uris_from_links",
    "document_uri_self_claim",
)
class TestDocumentURIsFromData(object):
    def test_it_gets_document_uris_from_links(self, document_uris_from_links):
        document_data = {
            "link": [
                # In production these would be link dicts not strings.
                "link_dict_1",
                "link_dict_2",
                "link_dict_3",
            ]
        }
        claimant = "http://localhost:5000/docs/help"
        document_uris_from_links.return_value = [
            {"uri": "uri_1"},
            {"uri": "uri_2"},
            {"uri": "uri_3"},
        ]

        document_uris = document_claims.document_uris_from_data(
            document_data=document_data, claimant=claimant
        )

        document_uris_from_links.assert_called_once_with(
            document_data["link"], claimant
        )
        for document_uri in document_uris_from_links.return_value:
            assert document_uri in document_uris

    def test_calling_document_uris_from_links_when_no_links(
        self, document_uris_from_links
    ):
        document_data = {}  # No 'link' key.
        claimant = "http://localhost:5000/docs/help"

        document_claims.document_uris_from_data(
            document_data=document_data, claimant=claimant
        )

        document_uris_from_links.assert_called_once_with([], claimant)

    def test_it_gets_documents_uris_from_highwire_pdf(
        self, document_uris_from_highwire_pdf
    ):
        document_data = {"highwire": {"pdf": ["pdf_1", "pdf_2", "pdf_3"]}}
        claimant = "http://localhost:5000/docs/help"
        document_uris_from_highwire_pdf.return_value = [
            {"uri": "uri_1"},
            {"uri": "uri_2"},
            {"uri": "uri_3"},
        ]

        document_uris = document_claims.document_uris_from_data(
            document_data=document_data, claimant=claimant
        )

        document_uris_from_highwire_pdf.assert_called_once_with(
            document_data["highwire"], claimant
        )
        for document_uri in document_uris_from_highwire_pdf.return_value:
            assert document_uri in document_uris

    def test_calling_document_uris_from_highwire_pdf_when_no_highwire(
        self, document_uris_from_highwire_pdf
    ):
        document_data = {}  # No 'highwire' key.
        claimant = "http://localhost:5000/docs/help"

        document_claims.document_uris_from_data(
            document_data=document_data, claimant=claimant
        )

        document_uris_from_highwire_pdf.assert_called_once_with({}, claimant)

    def test_it_gets_documents_uris_from_highwire_doi(
        self, document_uris_from_highwire_doi
    ):
        document_data = {"highwire": {"doi": ["doi_1", "doi_2", "doi_3"]}}
        claimant = "http://localhost:5000/docs/help"
        document_uris_from_highwire_doi.return_value = [
            {"uri": "uri_1"},
            {"uri": "uri_2"},
            {"uri": "uri_3"},
        ]

        document_uris = document_claims.document_uris_from_data(
            document_data=document_data, claimant=claimant
        )

        document_uris_from_highwire_doi.assert_called_once_with(
            document_data["highwire"], claimant
        )
        for document_uri in document_uris_from_highwire_doi.return_value:
            assert document_uri in document_uris

    def test_calling_document_uris_from_highwire_doi_when_no_highwire(
        self, document_uris_from_highwire_doi
    ):
        document_data = {}  # No 'highwire' key.
        claimant = "http://localhost:5000/docs/help"

        document_claims.document_uris_from_data(
            document_data=document_data, claimant=claimant
        )

        document_uris_from_highwire_doi.assert_called_once_with({}, claimant)

    def test_it_gets_documents_uris_from_dc(self, document_uris_from_dc):
        document_data = {"dc": {"identifier": ["doi_1", "doi_2", "doi_3"]}}
        claimant = "http://localhost:5000/docs/help"
        document_uris_from_dc.return_value = [
            {"uri": "uri_1"},
            {"uri": "uri_2"},
            {"uri": "uri_3"},
        ]

        document_uris = document_claims.document_uris_from_data(
            document_data=document_data, claimant=claimant
        )

        document_uris_from_dc.assert_called_once_with(document_data["dc"], claimant)
        for document_uri in document_uris_from_dc.return_value:
            assert document_uri in document_uris

    def test_calling_document_uris_from_dc_when_no_dc(self, document_uris_from_dc):
        document_data = {}  # No 'dc' key.
        claimant = "http://localhost:5000/docs/help"

        document_claims.document_uris_from_data(
            document_data=document_data, claimant=claimant
        )

        document_uris_from_dc.assert_called_once_with({}, claimant)

    def test_it_gets_self_claim_document_uris(self, document_uri_self_claim):
        claimant = "http://example.com/claimant"

        document_uris = document_claims.document_uris_from_data({}, claimant)

        document_uri_self_claim.assert_called_once_with(claimant)
        assert document_uri_self_claim.return_value in document_uris

    def test_it_ignores_null_uris(
        self,
        document_uris_from_links,
        document_uris_from_highwire_pdf,
        document_uris_from_highwire_doi,
        document_uris_from_dc,
        document_uri_self_claim,
    ):
        document_uris_from_links.return_value = [{"uri": None}]
        document_uris_from_highwire_pdf.return_value = [{"uri": None}]
        document_uris_from_highwire_doi.return_value = [{"uri": None}]
        document_uris_from_dc.return_value = [{"uri": None}]
        document_uri_self_claim.return_value = {"uri": None}

        document_uris = document_claims.document_uris_from_data(
            {}, "http://example.com/claimant"
        )

        assert document_uris == []

    def test_it_ignores_empty_string_uris(
        self,
        document_uris_from_links,
        document_uris_from_highwire_pdf,
        document_uris_from_highwire_doi,
        document_uris_from_dc,
        document_uri_self_claim,
    ):
        document_uris_from_links.return_value = [{"uri": ""}]
        document_uris_from_highwire_pdf.return_value = [{"uri": ""}]
        document_uris_from_highwire_doi.return_value = [{"uri": ""}]
        document_uris_from_dc.return_value = [{"uri": ""}]
        document_uri_self_claim.return_value = {"uri": ""}

        document_uris = document_claims.document_uris_from_data(
            {}, "http://example.com/claimant"
        )

        assert document_uris == []

    def test_it_ignores_whitespace_only_self_claim_uris(self, document_uri_self_claim):
        for uri in (" ", "\n ", "\r\n", " \t"):
            document_uri_self_claim.return_value = {"uri": uri}

            document_uris = document_claims.document_uris_from_data(
                {}, "http://example.com/claimant"
            )

            assert document_uris == []

    def test_it_ignores_whitespace_only_uris(
        self,
        document_uris_from_links,
        document_uris_from_highwire_pdf,
        document_uris_from_highwire_doi,
        document_uris_from_dc,
        document_uri_self_claim,
    ):
        uris = [" ", "\n ", "\r\n", " \t"]
        document_uris_from_links.return_value = [{"uri": u} for u in uris]
        document_uris_from_highwire_pdf.return_value = [{"uri": u} for u in uris]
        document_uris_from_highwire_doi.return_value = [{"uri": u} for u in uris]
        document_uris_from_dc.return_value = [{"uri": u} for u in uris]

        document_uris = document_claims.document_uris_from_data(
            {}, "http://example.com/claimant"
        )

        assert document_uris == [document_uri_self_claim.return_value]

    def test_it_strips_whitespace_from_uris(
        self,
        document_uris_from_links,
        document_uris_from_highwire_pdf,
        document_uris_from_highwire_doi,
        document_uris_from_dc,
        document_uri_self_claim,
        matchers,
    ):
        document_uris_from_links.return_value = [
            {"uri": " from_link_1"},
            {"uri": "from_link_2 "},
            {"uri": " from_link_3 "},
        ]
        document_uris_from_highwire_pdf.return_value = [
            {"uri": " highwire_1"},
            {"uri": "highwire_2 "},
            {"uri": " highwire_3 "},
        ]
        document_uris_from_highwire_doi.return_value = [
            {"uri": " doi_1"},
            {"uri": "doi_2 "},
            {"uri": " doi_3 "},
        ]
        document_uris_from_dc.return_value = [
            {"uri": " dc_1"},
            {"uri": "dc_2 "},
            {"uri": " dc_3 "},
        ]

        document_uris = document_claims.document_uris_from_data(
            {}, "http://example.com/claimant"
        )

        assert document_uris == matchers.UnorderedList(
            [
                {"uri": "from_link_1"},
                {"uri": "from_link_2"},
                {"uri": "from_link_3"},
                {"uri": "highwire_1"},
                {"uri": "highwire_2"},
                {"uri": "highwire_3"},
                {"uri": "doi_1"},
                {"uri": "doi_2"},
                {"uri": "doi_3"},
                {"uri": "dc_1"},
                {"uri": "dc_2"},
                {"uri": "dc_3"},
                document_uri_self_claim.return_value,
            ]
        )

    def test_it_strips_whitespace_from_self_claim_uris(
        self, document_uris_from_links, document_uri_self_claim
    ):
        for uri in (" self_claim", "self_claim ", " self_claim "):
            document_uris_from_links.return_value = []
            document_uri_self_claim.return_value = {"uri": uri}

            document_uris = document_claims.document_uris_from_data(
                {}, "http://example.com/claimant"
            )

            assert document_uris == [{"uri": uri.strip()}]

    @pytest.fixture
    def document_uris_from_dc(self, patch):
        return patch("hyputils.memex.util.document_claims.document_uris_from_dc", return_value=[])

    @pytest.fixture
    def document_uris_from_highwire_pdf(self, patch):
        return patch(
            "hyputils.memex.util.document_claims.document_uris_from_highwire_pdf", return_value=[]
        )

    @pytest.fixture
    def document_uris_from_highwire_doi(self, patch):
        return patch(
            "hyputils.memex.util.document_claims.document_uris_from_highwire_doi", return_value=[]
        )

    @pytest.fixture
    def document_uris_from_links(self, patch):
        return patch("hyputils.memex.util.document_claims.document_uris_from_links", return_value=[])

    @pytest.fixture
    def document_uri_self_claim(self, patch):
        return patch("hyputils.memex.util.document_claims.document_uri_self_claim")


def one(list_):
    assert len(list_) == 1
    return list_[0]


def strip_prefix(prefix, s):
    if s.startswith(prefix):
        return s[len(prefix) :]
    else:
        return s



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import datetime

import sqlalchemy as sa
from sqlalchemy.dialects import postgresql as pg
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.ext.mutable import MutableDict, MutableList

from hyputils.memex.db import Base, types
from hyputils.memex.util import markdown, uri
from hyputils.memex.util.user import split_user


class Annotation(Base):

    """Model class representing a single annotation."""

    __tablename__ = "annotation"
    __table_args__ = (
        # Tags are stored in an array-type column, and indexed using a
        # generalised inverted index. For more information on the use of GIN
        # indices for array columns, see:
        #
        #   http://www.databasesoup.com/2015/01/tag-all-things.html
        #   http://www.postgresql.org/docs/9.5/static/gin-intro.html
        #
        sa.Index("ix__annotation_tags", "tags", postgresql_using="gin"),
        sa.Index("ix__annotation_updated", "updated"),
        # This is a functional index on the *first* of the annotation's
        # references, pointing to the top-level annotation it refers to. We're
        # using 1 here because Postgres uses 1-based array indexing.
        sa.Index("ix__annotation_thread_root", sa.text('("references"[1])')),
    )

    #: Annotation ID: these are stored as UUIDs in the database, and mapped
    #: transparently to a URL-safe Base64-encoded string.
    id = sa.Column(
        types.URLSafeUUID, server_default=sa.func.uuid_generate_v1mc(), primary_key=True
    )

    #: The timestamp when the annotation was created.
    created = sa.Column(
        sa.DateTime,
        default=datetime.datetime.utcnow,
        server_default=sa.func.now(),
        nullable=False,
    )

    #: The timestamp when the user edited the annotation last.
    updated = sa.Column(
        sa.DateTime,
        server_default=sa.func.now(),
        default=datetime.datetime.utcnow,
        nullable=False,
    )

    #: The full userid (e.g. 'acct:foo@example.com') of the owner of this
    #: annotation.
    userid = sa.Column(sa.UnicodeText, nullable=False, index=True)
    #: The string id of the group in which this annotation is published.
    #: Defaults to the global public group, "__world__".
    groupid = sa.Column(
        sa.UnicodeText,
        default="__world__",
        server_default="__world__",
        nullable=False,
        index=True,
    )

    #: The textual body of the annotation.
    _text = sa.Column("text", sa.UnicodeText)
    #: The Markdown-rendered and HTML-sanitized textual body of the annotation.
    _text_rendered = sa.Column("text_rendered", sa.UnicodeText)

    #: The tags associated with the annotation.
    tags = sa.Column(
        MutableList.as_mutable(pg.ARRAY(sa.UnicodeText, zero_indexes=True))
    )

    #: A boolean indicating whether this annotation is shared with members of
    #: the group it is published in. "Private"/"Only me" annotations have
    #: shared=False.
    shared = sa.Column(
        sa.Boolean,
        nullable=False,
        default=False,
        server_default=sa.sql.expression.false(),
    )

    #: The URI of the annotated page, as provided by the client.
    _target_uri = sa.Column("target_uri", sa.UnicodeText)
    #: The URI of the annotated page in normalized form.
    _target_uri_normalized = sa.Column("target_uri_normalized", sa.UnicodeText)
    #: The serialized selectors for the annotation on the annotated page.
    target_selectors = sa.Column(
        types.AnnotationSelectorJSONB, default=list, server_default=sa.func.jsonb("[]")
    )

    #: An array of annotation IDs which are ancestors of this annotation.
    references = sa.Column(
        pg.ARRAY(types.URLSafeUUID, zero_indexes=True),
        default=list,
        server_default=sa.text("ARRAY[]::uuid[]"),
    )

    #: Any additional serialisable data provided by the client.
    extra = sa.Column(
        MutableDict.as_mutable(pg.JSONB),
        default=dict,
        server_default=sa.func.jsonb("{}"),
        nullable=False,
    )

    #: Has the annotation been deleted?
    deleted = sa.Column(
        sa.Boolean,
        nullable=False,
        default=False,
        server_default=sa.sql.expression.false(),
    )

    document_id = sa.Column(sa.Integer, sa.ForeignKey("document.id"), nullable=False)

    document = sa.orm.relationship("Document", backref="annotations")

    thread = sa.orm.relationship(
        "Annotation",
        primaryjoin=(sa.orm.foreign(id) == sa.orm.remote(references[0])),
        viewonly=True,
        uselist=True,
    )

    @hybrid_property
    def target_uri(self):
        return self._target_uri

    @target_uri.setter
    def target_uri(self, value):
        self._target_uri = value
        self._target_uri_normalized = uri.normalize(value)

    @hybrid_property
    def target_uri_normalized(self):
        return self._target_uri_normalized

    @hybrid_property
    def text(self):
        return self._text

    @text.setter
    def text(self, value):
        self._text = value
        # N.B. We MUST take care here of appropriately escaping the user
        # input. Code elsewhere will assume that the content of the
        # `text_rendered` field is safe for printing without further escaping.
        #
        # `markdown.render` does the hard work for now.
        self._text_rendered = markdown.render(value)

    @hybrid_property
    def text_rendered(self):
        return self._text_rendered

    @property
    def thread_ids(self):
        return [thread_annotation.id for thread_annotation in self.thread]

    @property
    def is_reply(self):
        return bool(self.references)

    @property
    def parent_id(self):
        """
        Return the ID of the annotation that this annotation is a reply to.

        Return None if this annotation is not a reply.

        """
        if self.references:
            return self.references[-1]

    @property
    def thread_root_id(self):
        """
        Return the ID of the root annotation of this annotation's thread.

        Return the ID of the root annotation of the thread to which this
        annotation belongs. May be this annotation's own ID if it is the root
        annotation of its thread.

        """
        if self.references:
            return self.references[0]
        else:
            return self.id

    @property
    def authority(self):
        """
        Return the authority of the user and group this annotation belongs to.

        For example, returns "hypothes.is" for Hypothesis first-party
        annotations, or "elifesciences.org" for eLife third-party annotations.

        If this annotation doesn't have a userid (which is possible for
        annotations that haven't been saved to the DB yet) then return None.

        :raises ValueError: if the annotation's userid is invalid

        """
        if self.userid is None:
            return None
        return split_user(self.userid)["domain"]

    def __repr__(self):
        return "<Annotation %s>" % self.id



import os
import unittest
import pytest
from hyputils.hypothesis import AnnoFetcher
from hyputils.hypothesis import api_token, username, group

get_annos = AnnoFetcher(api_token, username, group)


class TestFetch(unittest.TestCase):
    def test_no_limit(self):
        annos = get_annos.get_annos_from_api()
        assert len(annos) > 800

    def test_max_results(self):
        annos = get_annos.get_annos_from_api(max_results=400)

    def test_stop_at(self):
        _annos = get_annos.get_annos_from_api(max_results=400)

        in_ = _annos[236]
        sanity = _annos[:237]
        assert in_ == sanity[-1]
        assert len(sanity) == 237

        stop_at = _annos[236].updated
        annos = get_annos.get_annos_from_api(stop_at=stop_at)

        assert len(annos) == 237

    def test_search_after(self):
        annos = get_annos.get_annos_from_api(max_results=200)
        stop_at = annos[100].updated
        search_after = annos[99].updated
        annos = get_annos.get_annos_from_api(search_after=search_after, stop_at=stop_at)
        assert len(annos) == 1

    def test_mr_lower(self):
        annos = get_annos.get_annos_from_api(max_results=200)
        stop_at = annos[150].updated
        annos = get_annos.get_annos_from_api(stop_at=stop_at, max_results=100)
        assert len(annos) == 100

    def test_sa_lower(self):
        annos = get_annos.get_annos_from_api(max_results=50)
        stop_at = annos[-1].updated
        annos = get_annos.get_annos_from_api(stop_at=stop_at, max_results=100)
        print(len(annos))
        assert len(annos) == 50

    def test_bad_stop(self):
        try:
            annos = get_annos.get_annos_from_api(max_results=50)
            stop_at = annos[-1]  # passing annotaiton instead of updated
            annos = get_annos.get_annos_from_api(stop_at=stop_at, max_results=100)
            raise AssertionError('should fail due to a type error from annos[-1]')
        except TypeError:
            pass

    @pytest.mark.skipif('CI' not in os.environ, reason='CI test, test user must have no public annos')
    def test_world_nomax(self):
        bad_annos = AnnoFetcher(api_token, username, '__world__')
        annos = bad_annos.get_annos_from_api()
        assert not annos



# -*- coding: utf-8 -*-

"""
Matchers for use in tests.

This module contains objects which can be used to write tests which are easier
to read, especially when using assertions against :py:class:`mock.Mock`
objects. Each matcher defines an __eq__ method, allowing instances of these
matchers to be used wherever you would usually pass a plain value.

For example, imagine you wanted to ensure that the `set_value` method of the
`foo` mock object was called exactly once with an integer. Previously you
would have to do something like:

    assert foo.set_value.call_count == 1
    assert isinstance(foo.set_value.call_args[0][0], int)

By using the `InstanceOf` matcher you can simply write:

    foo.set_value.assert_called_once_with(matchers.InstanceOf(int))

As a bonus, the second test will print substantially more useful debugging
output if it fails, e.g.

    E       AssertionError: Expected call: set_value(<instance of <type 'int'>>)
    E       Actual call: set_value('a string')

"""
from __future__ import unicode_literals

import re


class Matcher(object):
    def __eq__(self, other):
        raise NotImplementedError("subclasses should provide an __eq__ method")

    def __ne__(self, other):
        return not self.__eq__(other)


class AnyCallable(Matcher):
    """An object __eq__ to any callable object."""

    def __eq__(self, other):
        """Return ``True`` if ``other`` is callable, ``False`` otherwise."""
        return callable(other)


class NativeString(Matcher):
    """
    Matches any native string with the given characters.

    "Native string" means the ``str`` type which is a byte string in Python 2
    and a unicode string in Python 3. Does not match unicode strings (type
    ``unicode``) in Python 2, or byte strings (type ``bytes``) in Python 3,
    even if they contain the same characters.

    In Python 3 a ``bytes`` is never ``==`` to a ``str`` anyway, even if they
    contain the same characters. But in Python 2 a ``str`` is equal to a
    ``unicode`` if they contain the same characters, and that's why this
    matcher is needed.

    TODO: Delete this matcher once we no longer support Python 2.

    """

    def __init__(self, string):
        self.string = str(string)

    def __eq__(self, other):
        if not isinstance(other, str):
            return False
        return other == self.string

    def __repr__(self):
        return '<native string matching "{string}">'.format(string=self.string)

    def lower(self):
        return NativeString(self.string.lower())


class InstanceOf(Matcher):
    """An object __eq__ to any object which is an instance of `type_`."""

    def __init__(self, type_):
        self.type = type_

    def __eq__(self, other):
        return isinstance(other, self.type)

    def __repr__(self):
        return "<instance of {!r}>".format(self.type)


class IterableWith(Matcher):
    """An object __eq__ to any iterable which yields `items`."""

    def __init__(self, items):
        self.items = items

    def __eq__(self, other):
        return list(other) == self.items

    def __repr__(self):
        return "<iterable with {!r}>".format(self.items)


class MappingContaining(Matcher):
    """An object __eq__ to any mapping with the passed `key`."""

    def __init__(self, key):
        self.key = key

    def __eq__(self, other):
        try:
            other[self.key]
        except (TypeError, KeyError):
            return False
        else:
            return True

    def __repr__(self):
        return "<mapping containing {!r}>".format(self.key)


class Regex(Matcher):
    """Matches any string matching the passed regex."""

    def __init__(self, patt):
        self.patt = re.compile(patt)

    def __eq__(self, other):
        return bool(self.patt.match(other))

    def __repr__(self):
        return "<string matching re {!r}>".format(self.patt.pattern)


class UnorderedList(Matcher):
    """
    Matches a list with the same items in any order.

    Matches any list that contains the same items as the given list
    (and no more), regardless of order.

    """

    def __init__(self, items):
        self.items = items

    def __eq__(self, other):
        if len(self.items) != len(other):
            return False
        for item in self.items:
            if item not in other:
                return False
        return True

    def __repr__(self):
        return "<unordered list containing {items}".format(items=self.items)



# -*- coding: utf-8 -*-
"""Classes for validating data passed to the annotations API."""
from __future__ import unicode_literals

import copy
from dateutil.parser import parse

from hyputils.memex.schemas.base import JSONSchema, ValidationError
from hyputils.memex.util import document_claims


class AnnotationSchema(JSONSchema):

    """Validate an annotation object."""

    schema = {
        "type": "object",
        "properties": {
            "document": {
                "type": "object",
                "properties": {
                    "dc": {
                        "type": "object",
                        "properties": {
                            "identifier": {"type": "array", "items": {"type": "string"}}
                        },
                    },
                    "highwire": {
                        "type": "object",
                        "properties": {
                            "doi": {"type": "array", "items": {"type": "string"}},
                            "pdf_url": {"type": "array", "items": {"type": "string"}},
                        },
                    },
                    "link": {
                        "type": "array",
                        "items": {
                            "type": "object",
                            "properties": {
                                "href": {"type": "string"},
                                "type": {"type": "string"},
                            },
                            "required": ["href"],
                        },
                    },
                },
            },
            "group": {"type": "string"},
            "permissions": {
                "title": "Permissions",
                "description": "Annotation action access control list",
                "type": "object",
                "patternProperties": {
                    "^(admin|delete|read|update)$": {
                        "type": "array",
                        "items": {"type": "string", "pattern": "^(acct:|group:).+$"},
                    }
                },
                "required": ["read"],
            },
            "references": {"type": "array", "items": {"type": "string"}},
            "tags": {"type": "array", "items": {"type": "string"}},
            "target": {
                "type": "array",
                "items": {
                    "type": "object",
                    "properties": {
                        "selector": {
                            "type": "array",
                            "items": {
                                "type": "object",
                                "properties": {"type": {"type": "string"}},
                                "required": ["type"],
                            },
                        }
                    },
                },
            },
            "text": {"type": "string"},
            "uri": {"type": "string"},
        },
    }


class CreateAnnotationSchema(object):

    """Validate the POSTed data of a create annotation request."""

    def __init__(self, request):
        self.structure = AnnotationSchema()
        self.request = request

    def validate(self, data, remove_protected=True):
        appstruct = self.structure.validate(data)

        new_appstruct = {}

        if remove_protected:
            _remove_protected_fields(appstruct)

        new_appstruct["userid"] = self.request.authenticated_userid

        uri = appstruct.pop("uri", "").strip()
        if not uri:
            raise ValidationError("uri: " + "'uri' is a required property")
        new_appstruct["target_uri"] = uri

        new_appstruct["text"] = appstruct.pop("text", "")
        new_appstruct["tags"] = appstruct.pop("tags", [])
        new_appstruct["groupid"] = appstruct.pop("group", "__world__")
        new_appstruct["references"] = appstruct.pop("references", [])

        if "permissions" in appstruct:
            new_appstruct["shared"] = _shared(
                appstruct.pop("permissions"), new_appstruct["groupid"]
            )
        else:
            new_appstruct["shared"] = False

        if "target" in appstruct:
            new_appstruct["target_selectors"] = _target_selectors(
                appstruct.pop("target")
            )

        # Replies always get the same groupid as their parent. The parent's
        # groupid is added to the reply annotation later by the storage code.
        # Here we just delete any group sent by the client from replies.
        if new_appstruct["references"] and "groupid" in new_appstruct:
            del new_appstruct["groupid"]

        new_appstruct["document"] = _document(
            appstruct.pop("document", {}), new_appstruct["target_uri"]
        )

        new_appstruct["extra"] = appstruct

        return new_appstruct


class UpdateAnnotationSchema(object):

    """Validate the POSTed data of an update annotation request."""

    def __init__(self, request, existing_target_uri, groupid):
        self.request = request
        self.existing_target_uri = existing_target_uri
        self.groupid = groupid
        self.structure = AnnotationSchema()

    def validate(self, data, remove_protected=True):
        appstruct = self.structure.validate(data)

        new_appstruct = {}

        if remove_protected:
            _remove_protected_fields(appstruct)

        # Some fields are not allowed to be changed in annotation updates.
        for key in ["group", "groupid", "userid", "references"]:
            appstruct.pop(key, "")

        # Fields that are allowed to be updated and that have a different name
        # internally than in the public API.
        if "uri" in appstruct:
            new_uri = appstruct.pop("uri").strip()
            if not new_uri:
                raise ValidationError("uri: " + "'uri' is a required property")
            new_appstruct["target_uri"] = new_uri

        if "permissions" in appstruct:
            new_appstruct["shared"] = _shared(
                appstruct.pop("permissions"), self.groupid
            )

        if "target" in appstruct:
            new_appstruct["target_selectors"] = _target_selectors(
                appstruct.pop("target")
            )

        # Fields that are allowed to be updated and that have the same internal
        # and external name.
        for key in ["text", "tags"]:
            if key in appstruct:
                new_appstruct[key] = appstruct.pop(key)

        if "document" in appstruct:
            new_appstruct["document"] = _document(
                appstruct.pop("document"),
                new_appstruct.get("target_uri", self.existing_target_uri),
            )

        new_appstruct["extra"] = appstruct

        return new_appstruct


def _document(document, claimant):
    """
    Return document meta and document URI data from the given document dict.

    Transforms the "document" dict that the client posts into a convenient
    format for creating DocumentURI and DocumentMeta objects later.

    """
    document = document or {}
    document_uri_dicts = document_claims.document_uris_from_data(
        copy.deepcopy(document), claimant=claimant
    )
    document_meta_dicts = document_claims.document_metas_from_data(
        copy.deepcopy(document), claimant=claimant
    )
    return {
        "document_uri_dicts": document_uri_dicts,
        "document_meta_dicts": document_meta_dicts,
    }


def _format_jsonschema_error(error):
    """Format a :py:class:`jsonschema.ValidationError` as a string."""
    if error.path:
        dotted_path = ".".join([str(c) for c in error.path])
        return "{path}: {message}".format(path=dotted_path, message=error.message)
    return error.message


def _remove_protected_fields(appstruct):
    # Some fields are not to be set by the user, ignore them.
    for field in [
        "created",
        "updated",
        "user",
        "id",
        "links",
        "flagged",
        "hidden",
        "moderation",
        "user_info",
    ]:
        appstruct.pop(field, None)


def _shared(permissions, groupid):
    """
    Return True if the given permissions object represents shared permissions.

    Return False otherwise.

    Reduces the client's complex permissions dict to a simple shared boolean.

    :param permissions: the permissions dict sent by the client in an
        annotation create or update request
    :type permissions: dict

    :param groupid: the groupid of the annotation that the permissions dict
        applies to
    :type groupid: unicode

    """
    return permissions["read"] == ["group:{id}".format(id=groupid)]


def _target_selectors(targets):
    """
    Return the target selectors from the given target list.

    Transforms the target lists that the client sends in annotation create and
    update requests into our internal target_selectors format.

    """
    # Any targets other than the first in the list are discarded.
    # Any fields of the target other than 'selector' are discarded.
    if targets and "selector" in targets[0]:
        return targets[0]["selector"]
    else:
        return []




# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import sqlalchemy as sa

from hyputils.memex.db import Base


class UserIdentity(Base):
    __tablename__ = "user_identity"
    __table_args__ = (sa.UniqueConstraint("provider", "provider_unique_id"),)

    id = sa.Column(sa.Integer(), autoincrement=True, primary_key=True)
    provider = sa.Column(sa.UnicodeText(), nullable=False)
    provider_unique_id = sa.Column(sa.UnicodeText(), nullable=False)
    user_id = sa.Column(
        sa.Integer(), sa.ForeignKey("user.id", ondelete="cascade"), nullable=False
    )

    def __repr__(self):
        return "{}(provider={!r}, provider_unique_id={!r})".format(
            self.__class__.__name__, self.provider, self.provider_unique_id
        )



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import pytest

from hyputils.memex.models.annotation import Annotation


def test_parent_id_of_direct_reply():
    ann = Annotation(references=["parent_id"])

    assert ann.parent_id == "parent_id"


def test_parent_id_of_reply_to_reply():
    ann = Annotation(references=["reply1", "reply2", "parent_id"])

    assert ann.parent_id == "parent_id"


def test_reply_is_reply():
    ann = Annotation(references=["parent_id"])

    assert ann.is_reply is True


def test_non_reply_is_not_reply():
    ann = Annotation()

    assert ann.is_reply is False


def test_parent_id_of_annotation():
    ann = Annotation()

    assert ann.parent_id is None


def test_thread_root_id_returns_id_if_no_references():
    annotation = Annotation(id="GBhy1DoHEea6htPothzqZQ")

    assert annotation.thread_root_id == "GBhy1DoHEea6htPothzqZQ"


def test_thread_root_id_returns_id_if_references_empty():
    annotation = Annotation(id="jANlljoHEea6hsv8FY7ipw", references=[])

    assert annotation.thread_root_id == "jANlljoHEea6hsv8FY7ipw"


def test_thread_root_id_returns_reference_if_only_one_reference():
    annotation = Annotation(
        id="qvJnIjoHEea6hiv0nJK7gw", references=["yiSVIDoHEea6hjcSFuROLw"]
    )

    assert annotation.thread_root_id == "yiSVIDoHEea6hjcSFuROLw"


def test_thread_root_id_returns_first_reference_if_many_references():
    annotation = Annotation(
        id="uK9yVjoHEea6hsewWuiKtQ",
        references=[
            "1Ife3DoHEea6hpv8vWujdQ",
            "uVuItjoHEea6hiNgv1wvmg",
            "Qe7fpc5ZRgWy0RSHEP9UNg",
        ],
    )

    assert annotation.thread_root_id == "1Ife3DoHEea6hpv8vWujdQ"


def test_text_setter_renders_markdown(markdown):
    markdown.render.return_value = "<p>foobar</p>"

    annotation = Annotation()
    annotation.text = "foobar"

    markdown.render.assert_called_once_with("foobar")

    annotation.text_rendered == markdown.render.return_value


@pytest.mark.parametrize(
    "userid,authority",
    [
        ("acct:bmason@hypothes.is", "hypothes.is"),
        ("acct:kaylawatson@elifesciences.org", "elifesciences.org"),
    ],
)
def test_authority(factories, userid, authority):
    assert factories.Annotation(userid=userid).authority == authority


def test_authority_when_annotation_has_no_userid():
    assert Annotation().authority is None


def test_setting_extras_inline_is_persisted(db_session, factories):
    """
    In-place changes to Annotation.extra should be persisted.

    Setting an Annotation.extra value in-place:

        my_annotation.extra['foo'] = 'bar'

    should be persisted to the database.

    """
    annotation = factories.Annotation(userid="fred")

    annotation.extra["foo"] = "bar"

    # We need to commit the db session here so that the in-place change to
    # annotation.extra above would be lost if annotation.extra was a normal
    # dict. Without this commit() this test would never fail.
    db_session.commit()

    annotation = db_session.query(Annotation).get(annotation.id)

    assert annotation.extra == {"foo": "bar"}


def test_deleting_extras_inline_is_persisted(db_session, factories):
    """
    In-place changes to Annotation.extra should be persisted.

    Deleting an Annotation.extra value in-place should be persisted to the
    database.

    """
    annotation = factories.Annotation(userid="fred", extra={"foo": "bar"})

    del annotation.extra["foo"]
    db_session.commit()
    annotation = db_session.query(Annotation).get(annotation.id)

    assert "foo" not in annotation.extra


def test_appending_tags_inline_is_persisted(db_session, factories):
    """
    In-place changes to Annotation.tags should be persisted.

    Changes made by Annotation.tags.append() should be persisted to the
    database.

    """
    annotation = factories.Annotation(userid="fred", tags=["foo"])

    annotation.tags.append("bar")
    db_session.commit()
    annotation = db_session.query(Annotation).get(annotation.id)

    assert "bar" in annotation.tags


def test_deleting_tags_inline_is_persisted(db_session, factories):
    """In-place deletions of annotation tags should be persisted."""
    annotation = factories.Annotation(userid="fred", tags=["foo"])

    del annotation.tags[0]
    db_session.commit()
    annotation = db_session.query(Annotation).get(annotation.id)

    assert "foo" not in annotation.tags


class TestThread(object):
    def test_empty_thread(self, root):
        assert root.thread == []

    def test_empty_thread_ids(self, root):
        assert root.thread_ids == []

    def test_thread_with_replies(self, root, reply, subreply):
        assert set(root.thread) == set([reply, subreply])

    def test_thread_ids_with_replies(self, root, reply, subreply):
        assert set(root.thread_ids) == set([reply.id, subreply.id])

    @pytest.mark.usefixtures("subreply")
    def test_reply_has_no_thread(self, reply):
        assert reply.thread == []

    @pytest.mark.usefixtures("subreply")
    def test_reply_has_no_thread_ids(self, reply):
        assert reply.thread_ids == []

    @pytest.fixture
    def root(self, factories):
        return factories.Annotation()

    @pytest.fixture
    def reply(self, factories, root):
        return factories.Annotation(references=[root.id])

    @pytest.fixture
    def subreply(self, factories, root, reply):
        return factories.Annotation(references=[root.id, reply.id])


@pytest.fixture
def markdown(patch):
    return patch("hyputils.memex.models.annotation.markdown")



class AllPermissionsList(object):
    """ Stand in 'permission list' to represent all permissions """
    def __iter__(self):
        return ()
    def __contains__(self, other):
        return True
    def __eq__(self, other):
        return isinstance(other, self.__class__)


ALL_PERMISSIONS = AllPermissionsList()


def is_nonstr_iter(v):
    """ from pyramid.compat """
    if isinstance(v, str):
        return False
    return hasattr(v, '__iter__')


def lineage(resource):
    """ from pyramid.location """
    while resource is not None:
        yield resource
        try:
            resource = resource.__parent__
        except AttributeError:
            resource = None


class security:
    """ local version of pyramid security object """
    Allow = 'Allow'
    Deny = 'Deny'
    Authenticated = 'system.Authenticated'
    Everyone = 'system.Everyone'
    DENY_ALL = ('Deny', 'system.Everyone', AllPermissionsList())
    ACLDenied = 0
    ACLAllowed = 1
    class ACLAuthorizationPolicy(object):
        """ stripped down acl auth policy """

        @staticmethod
        def permits(context, principals, permission):

            acl = '<No ACL found on any object in resource lineage>'

            for location in lineage(context):
                try:
                    acl = location.__acl__
                except AttributeError:
                    continue

                if acl and callable(acl):
                    acl = acl()

                for ace in acl:
                    ace_action, ace_principal, ace_permissions = ace
                    if ace_principal in principals:
                        if not is_nonstr_iter(ace_permissions):
                            ace_permissions = [ace_permissions]
                        if permission in ace_permissions:
                            if ace_action == security.Allow:
                                return security.ACLAllowed
                            else:
                                return security.ACLDenied

            return security.ACLDenied

        def principals_allowed_by_permission(self, context, permission):
            allowed = set()

            for location in reversed(list(lineage(context))):
                # NB: we're walking *up* the object graph from the root
                try:
                    acl = location.__acl__
                except AttributeError:
                    continue

                allowed_here = set()
                denied_here = set()

                if acl and callable(acl):
                    acl = acl()

                for ace_action, ace_principal, ace_permissions in acl:
                    if not is_nonstr_iter(ace_permissions):
                        ace_permissions = [ace_permissions]
                    if (ace_action == security.Allow) and (permission in ace_permissions):
                        if not ace_principal in denied_here:
                            allowed_here.add(ace_principal)
                    if (ace_action == security.Deny) and (permission in ace_permissions):
                            denied_here.add(ace_principal)
                            if ace_principal == security.Everyone:
                                # clear the entire allowed set, as we've hit a
                                # deny of Everyone ala (Deny, Everyone, ALL)
                                allowed = set()
                                break
                            elif ace_principal in allowed:
                                allowed.remove(ace_principal)

                allowed.update(allowed_here)

            return allowed



import unittest
from hyputils.hypothesis import Memoizer, HypothesisHelper
from hyputils.hypothesis import api_token, username, group, group_to_memfile

get_annos = Memoizer(group_to_memfile(group), api_token, username, group)

annos = get_annos()


class TestZHelper(unittest.TestCase):
    """ NOTE these tests are stateful """

    def test_0_partial(self):
        partial = [HypothesisHelper(a, annos) for a in annos[:100]]
        repr(partial)
        assert len(partial) == 100

    def test_1_helper(self):
        [HypothesisHelper(a, annos) for a in annos[100:]]
        hh = list(HypothesisHelper)
        assert len(hh) > 800
        repr(hh)

    def test_2_by_id(self):
        h = next(iter(HypothesisHelper))
        hanno = HypothesisHelper.byId(h.id)
        assert hanno == h  # that silent failure on passing in h instead of h.id though ...

    def test_3_by_id_missing(self):
        hanno = HypothesisHelper.byId('not a real id')
        assert hanno is None

    def test_4_populate(self):
        [h.populateTags() for h in HypothesisHelper]

    def test_5_by_tags(self):
        test_tag = None
        max_tags = 0
        for tag, tset in HypothesisHelper._tagIndex.items():
            lt = len(tset)
            if lt > max_tags:
                max_tags = lt
                test_tag = tag
        hannos = HypothesisHelper.byTags(test_tag)
        assert all(test_tag in h.tags for h in hannos)

    def test_6_not_annos(self):
        HypothesisHelper(annos[0], annos[:10])



import unittest
from hyputils.hypothesis import HypothesisUtils
from hyputils.hypothesis import api_token, username, group

h = HypothesisUtils(username, api_token, group)


class TestCreateAnno(unittest.TestCase):
    def test_payload(self):
        kwargs = (('url', 'this is a url i swear'),
                  ('prefix', 'some prefix text'),
                  ('exact', 'my exact text'),
                  ('suffix', 'some suffix text'),
                  ('text', 'angry words being written about the exact text'),
                  ('tags', 'TEST:WHYHAVEYOUDONETHIS'),
                  ('document', None),
                  ('extra', None),)
        args = tuple(b for a, b in kwargs)
        payload = h.make_annotation_payload_with_target_using_only_text_quote(*args)


class TestHTTP(unittest.TestCase):
    anno_id = 'lCCu3LNFEeiAz7v-JjOpjQ'
    tags = ['RRID:AB_303684']

    def test_head(self):
        resp = h.head_annotation(self.anno_id)

    def test_get(self):
        resp = h.get_annotation(self.anno_id)
        assert resp.json()['tags'] == self.tags



# -*- coding: utf-8 -*-

"""Shared functionality for schemas."""

from __future__ import unicode_literals

import copy

import jsonschema


class ValidationError(Exception):
    pass


class JSONSchema(object):
    """
    Validate data according to a Draft 4 JSON Schema.

    Inherit from this class and override the `schema` class property with a
    valid JSON schema.
    """

    schema = {}

    def __init__(self):
        format_checker = jsonschema.FormatChecker()
        self.validator = jsonschema.Draft4Validator(
            self.schema, format_checker=format_checker
        )

    def validate(self, data):
        """
        Validate `data` according to the current schema.

        :param data: The data to be validated
        :returns: valid data
        :raises ~h.schemas.ValidationError: if the data is invalid
        """
        # Take a copy to ensure we don't modify what we were passed.
        appstruct = copy.deepcopy(data)

        errors = list(self.validator.iter_errors(appstruct))
        if errors:
            msg = ", ".join([_format_jsonschema_error(e) for e in errors])
            raise ValidationError(msg)
        return appstruct


def _format_jsonschema_error(error):
    """Format a :py:class:`jsonschema.ValidationError` as a string."""
    if error.path:
        dotted_path = ".".join([str(c) for c in error.path])
        return "{path}: {message}".format(path=dotted_path, message=error.message)
    return error.message



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from copy import deepcopy
import mock
import pytest
import re
from webob.multidict import NestedMultiDict, MultiDict

from hyputils.memex.schemas import ValidationError
from hyputils.memex.schemas.annotation import (
    CreateAnnotationSchema,
    UpdateAnnotationSchema,
)


def create_annotation_schema_validate(request, data):
    # 'uri' is required when creating new annotations.
    if "uri" not in data:
        data["uri"] = "http://example.com/example"

    schema = CreateAnnotationSchema(request)
    return schema.validate(data)


def update_annotation_schema_validate(
    request, data, existing_target_uri="", groupid=""
):
    schema = UpdateAnnotationSchema(request, existing_target_uri, groupid)
    return schema.validate(data)


@pytest.mark.parametrize(
    "validate",
    [create_annotation_schema_validate, update_annotation_schema_validate],
    ids=["CreateAnnotationSchema.validate()", "UpdateAnnotationSchema.validate()"],
)
class TestCreateUpdateAnnotationSchema(object):

    """Shared tests for CreateAnnotationSchema and UpdateAnnotationSchema."""

    def test_it_does_not_raise_for_minimal_valid_data(self, pyramid_request, validate):
        validate(pyramid_request, {})

    def test_it_does_not_raise_for_full_valid_data(self, pyramid_request, validate):
        # Use all the keys to make sure that valid data for all of them passes.
        validate(
            pyramid_request,
            {
                "document": {
                    "dc": {"identifier": ["foo", "bar"]},
                    "highwire": {"doi": ["foo", "bar"], "pdf_url": ["foo", "bar"]},
                    "link": [
                        {"href": "foo", "type": "foo"},
                        {"href": "foo", "type": "foo"},
                    ],
                },
                "group": "foo",
                "permissions": {
                    "admin": ["acct:foo", "group:bar"],
                    "delete": ["acct:foo", "group:bar"],
                    "read": ["acct:foo", "group:bar"],
                    "update": ["acct:foo", "group:bar"],
                },
                "references": ["foo", "bar"],
                "tags": ["foo", "bar"],
                "target": [
                    {"selector": [{"type": "foo"}]},
                    {"selector": [{"type": "bar"}]},
                ],
                "text": "foo",
                "uri": "foo",
            },
        )

    @pytest.mark.parametrize(
        "input_data,error_message",
        [
            ({"document": False}, "document: False is not of type 'object'"),
            ({"document": {"dc": False}}, "document.dc: False is not of type 'object'"),
            (
                {"document": {"dc": {"identifier": False}}},
                "document.dc.identifier: False is not of type 'array'",
            ),
            (
                {"document": {"dc": {"identifier": [False]}}},
                "document.dc.identifier.0: False is not of type 'string'",
            ),
            (
                {"document": {"highwire": False}},
                "document.highwire: False is not of type 'object'",
            ),
            (
                {"document": {"highwire": {"doi": False}}},
                "document.highwire.doi: False is not of type 'array'",
            ),
            (
                {"document": {"highwire": {"doi": [False]}}},
                "document.highwire.doi.0: False is not of type 'string'",
            ),
            (
                {"document": {"highwire": {"pdf_url": False}}},
                "document.highwire.pdf_url: False is not of type 'array'",
            ),
            (
                {"document": {"highwire": {"pdf_url": [False]}}},
                "document.highwire.pdf_url.0: False is not of type 'string'",
            ),
            (
                {"document": {"link": False}},
                "document.link: False is not of type 'array'",
            ),
            (
                {"document": {"link": [False]}},
                "document.link.0: False is not of type 'object'",
            ),
            (
                {"document": {"link": [{}]}},
                "document.link.0: 'href' is a required property",
            ),
            (
                {"document": {"link": [{"href": False}]}},
                "document.link.0.href: False is not of type 'string'",
            ),
            (
                {"document": {"link": [{"href": "http://example.com", "type": False}]}},
                "document.link.0.type: False is not of type 'string'",
            ),
            ({"group": False}, "group: False is not of type 'string'"),
            ({"permissions": False}, "permissions: False is not of type 'object'"),
            ({"permissions": {}}, "permissions: 'read' is a required property"),
            (
                {"permissions": {"read": False}},
                "permissions.read: False is not of type 'array'",
            ),
            (
                {"permissions": {"read": [False]}},
                "permissions.read.0: False is not of type 'string'",
            ),
            (
                {"permissions": {"read": ["foo"]}},
                "permissions.read.0: 'foo' does not match '^(acct:|group:).+$'",
            ),
            ({"references": False}, "references: False is not of type 'array'"),
            ({"references": [False]}, "references.0: False is not of type 'string'"),
            ({"tags": False}, "tags: False is not of type 'array'"),
            ({"tags": [False]}, "tags.0: False is not of type 'string'"),
            ({"target": False}, "target: False is not of type 'array'"),
            ({"target": [False]}, "target.0: False is not of type 'object'"),
            (
                {"target": [{"selector": False}]},
                "target.0.selector: False is not of type 'array'",
            ),
            (
                {"target": [{"selector": {"type": "foo"}}]},
                "target.0.selector: {'type': 'foo'} is not of type 'array'",
            ),
            (
                {"target": [{"selector": [False]}]},
                "target.0.selector.0: False is not of type 'object'",
            ),
            (
                {"target": [{"selector": [{}]}]},
                "target.0.selector.0: 'type' is a required property",
            ),
            (
                {"target": [{"selector": [{"type": False}]}]},
                "target.0.selector.0.type: False is not of type 'string'",
            ),
            ({"text": False}, "text: False is not of type 'string'"),
            ({"uri": False}, "uri: False is not of type 'string'"),
            ({"uri": ""}, "uri: 'uri' is a required property"),
            ({"uri": " "}, "uri: 'uri' is a required property"),
        ],
    )
    def test_it_raises_for_invalid_data(
        self, pyramid_request, validate, input_data, error_message
    ):
        with pytest.raises(ValidationError) as exc:
            validate(pyramid_request, input_data)

        # Strip "u" literal prefixes that get added in front of property names in
        # certain error messages in Python 2.
        actual_msg = str(exc.value)
        actual_msg = re.sub(r"u'([^']+)'", "'\\1'", actual_msg)
        assert actual_msg == error_message

    @pytest.mark.parametrize(
        "field",
        [
            "created",
            "updated",
            "user",
            "id",
            "links",
            "flagged",
            "hidden",
            "moderation",
            "user_info",
        ],
    )
    def test_it_removes_protected_fields(self, pyramid_request, validate, field):
        data = {}
        data[field] = "something forbidden"
        appstruct = validate(pyramid_request, data)

        assert field not in appstruct
        assert field not in appstruct.get("extra", {})

    def test_it_renames_uri_to_target_uri(self, pyramid_request, validate):
        appstruct = validate(pyramid_request, {"uri": "http://example.com/example"})

        assert appstruct["target_uri"] == "http://example.com/example"
        assert "uri" not in appstruct

    def test_it_strips_leading_and_trailing_whitespace_from_uri(
        self, pyramid_request, validate
    ):
        appstruct = validate(pyramid_request, {"uri": " foo "})

        assert appstruct["target_uri"] == "foo"

    def test_it_keeps_text(self, pyramid_request, validate):
        appstruct = validate(pyramid_request, {"text": "some annotation text"})

        assert appstruct["text"] == "some annotation text"

    def test_it_keeps_tags(self, pyramid_request, validate):
        appstruct = validate(pyramid_request, {"tags": ["foo", "bar"]})

        assert appstruct["tags"] == ["foo", "bar"]

    def test_it_replaces_target_with_target_selectors(self, pyramid_request, validate):
        appstruct = validate(
            pyramid_request,
            {
                "target": [
                    {
                        "foo": "bar",  # This should be removed
                        "selector": [{"type": "FooSelector"}, {"type": "BarSelector"}],
                    },
                    {
                        # Additional targets are ignored.
                        "selector": [{"type": "BazSelector"}]
                    },
                ]
            },
        )

        assert appstruct["target_selectors"] == [
            {"type": "FooSelector"},
            {"type": "BarSelector"},
        ]

    def test_it_extracts_document_uris_from_the_document(
        self, pyramid_request, document_claims, validate
    ):
        target_uri = "http://example.com/example"
        document_data = {"foo": "bar"}

        validate(pyramid_request, {"document": document_data, "uri": target_uri})

        document_claims.document_uris_from_data.assert_called_once_with(
            document_data, claimant=target_uri
        )

    def test_it_puts_document_uris_in_appstruct(
        self, document_claims, pyramid_request, validate
    ):
        appstruct = validate(pyramid_request, {"document": {}})

        assert appstruct["document"]["document_uri_dicts"] == (
            document_claims.document_uris_from_data.return_value
        )

    def test_it_extracts_document_metas_from_the_document(
        self, document_claims, pyramid_request, validate
    ):
        document_data = {"foo": "bar"}
        target_uri = "http://example.com/example"

        validate(pyramid_request, {"document": {"foo": "bar"}, "uri": target_uri})

        document_claims.document_metas_from_data.assert_called_once_with(
            document_data, claimant=target_uri
        )

    def test_it_does_not_pass_modified_dict_to_document_metas_from_data(
        self, document_claims, pyramid_request, validate
    ):
        """

        If document_uris_from_data() modifies the document dict that it's
        given, the original dict (or one with the same values as it) should be
        passed t document_metas_from_data(), not the modified copy.

        """
        document = {
            "top_level_key": "original_value",
            "sub_dict": {"key": "original_value"},
        }

        def document_uris_from_data(document, claimant):
            document["new_key"] = "new_value"
            document["top_level_key"] = "new_value"
            document["sub_dict"]["key"] = "new_value"

        document_claims.document_uris_from_data.side_effect = document_uris_from_data

        validate(pyramid_request, {"document": document})

        assert document_claims.document_metas_from_data.call_args[0][0] == document

    def test_it_puts_document_metas_in_appstruct(
        self, document_claims, pyramid_request, validate
    ):
        appstruct = validate(pyramid_request, {"document": {}})

        assert appstruct["document"]["document_meta_dicts"] == (
            document_claims.document_metas_from_data.return_value
        )

    def test_it_clears_existing_keys_from_document(self, pyramid_request, validate):
        """
        Any keys in the document dict should be removed.

        They're replaced with the 'document_uri_dicts' and
        'document_meta_dicts' keys.

        """
        appstruct = validate(
            pyramid_request, {"document": {"foo": "bar"}}  # This should be deleted.
        )

        assert "foo" not in appstruct["document"]

    def test_document_does_not_end_up_in_extra(self, pyramid_request, validate):
        appstruct = validate(pyramid_request, {"document": {"foo": "bar"}})

        assert "document" not in appstruct.get("extra", {})

    def test_it_moves_extra_data_into_extra_sub_dict(self, pyramid_request, validate):
        appstruct = validate(
            pyramid_request,
            {
                # Throw in all the fields, just to make sure that none of them get
                # into extra.
                "created": "created",
                "updated": "updated",
                "user": "user",
                "id": "id",
                "uri": "uri",
                "text": "text",
                "tags": ["gar", "har"],
                "permissions": {"read": ["group:__world__"]},
                "target": [],
                "group": "__world__",
                "references": ["parent"],
                "flagged": True,
                # These should end up in extra.
                "foo": 1,
                "bar": 2,
            },
        )

        assert appstruct["extra"] == {"foo": 1, "bar": 2}

    def test_it_does_not_modify_extra_fields_that_are_not_sent(
        self, pyramid_request, validate
    ):
        appstruct = validate(pyramid_request, {"foo": "bar"})

        assert "custom" not in appstruct["extra"]

    def test_it_does_not_modify_extra_fields_if_none_are_sent(
        self, pyramid_request, validate
    ):
        appstruct = validate(pyramid_request, {})

        assert not appstruct.get("extra")


class TestCreateAnnotationSchema(object):
    def test_it_raises_if_data_has_no_uri(self, pyramid_request):
        data = self.valid_data()
        del data["uri"]
        schema = CreateAnnotationSchema(pyramid_request)

        with pytest.raises(ValidationError) as exc:
            schema.validate(data)

        assert str(exc.value) == "uri: 'uri' is a required property"

    def test_it_inserts_empty_string_if_data_contains_no_text(self, pyramid_request):
        schema = CreateAnnotationSchema(pyramid_request)

        assert schema.validate(self.valid_data())["text"] == ""

    def test_it_inserts_empty_list_if_data_contains_no_tags(self, pyramid_request):
        schema = CreateAnnotationSchema(pyramid_request)

        assert schema.validate(self.valid_data())["tags"] == []

    def test_it_replaces_private_permissions_with_shared_False(self, pyramid_request):
        schema = CreateAnnotationSchema(pyramid_request)

        appstruct = schema.validate(
            self.valid_data(permissions={"read": ["acct:harriet@example.com"]})
        )

        assert appstruct["shared"] is False
        assert "permissions" not in appstruct

    def test_it_replaces_shared_permissions_with_shared_True(self, pyramid_request):
        schema = CreateAnnotationSchema(pyramid_request)

        appstruct = schema.validate(
            self.valid_data(
                permissions={"read": ["group:__world__"]}, group="__world__"
            )
        )

        assert appstruct["shared"] is True
        assert "permissions" not in appstruct

    def test_it_defaults_to_private_if_no_permissions_object_sent(
        self, pyramid_request
    ):
        schema = CreateAnnotationSchema(pyramid_request)

        appstruct = schema.validate(self.valid_data())

        assert appstruct["shared"] is False

    def test_it_renames_group_to_groupid(self, pyramid_request):
        schema = CreateAnnotationSchema(pyramid_request)

        appstruct = schema.validate(self.valid_data(group="foo"))

        assert appstruct["groupid"] == "foo"
        assert "group" not in appstruct

    def test_it_inserts_default_groupid_if_no_group(self, pyramid_request):
        schema = CreateAnnotationSchema(pyramid_request)

        appstruct = schema.validate(self.valid_data())

        assert appstruct["groupid"] == "__world__"

    def test_it_keeps_references(self, pyramid_request):
        schema = CreateAnnotationSchema(pyramid_request)

        appstruct = schema.validate(
            self.valid_data(references=["parent id", "parent id 2"])
        )

        assert appstruct["references"] == ["parent id", "parent id 2"]

    def test_it_inserts_empty_list_if_no_references(self, pyramid_request):
        schema = CreateAnnotationSchema(pyramid_request)

        appstruct = schema.validate(self.valid_data())

        assert appstruct["references"] == []

    def test_it_deletes_groupid_for_replies(self, pyramid_request):
        schema = CreateAnnotationSchema(pyramid_request)

        appstruct = schema.validate(
            self.valid_data(group="foo", references=["parent annotation id"])
        )

        assert "groupid" not in appstruct

    def valid_data(self, **kwargs):
        """Return minimal valid data for creating a new annotation."""
        data = {"uri": "http://example.com/example"}
        data.update(kwargs)
        return data


class TestUpdateAnnotationSchema(object):
    def test_you_cannot_change_an_annotations_group(self, pyramid_request):
        schema = UpdateAnnotationSchema(pyramid_request, "", "")

        appstruct = schema.validate({"groupid": "new-group", "group": "new-group"})

        assert "groupid" not in appstruct
        assert "groupid" not in appstruct.get("extra", {})
        assert "group" not in appstruct
        assert "group" not in appstruct.get("extra", {})

    def test_you_cannot_change_an_annotations_userid(self, pyramid_request):
        schema = UpdateAnnotationSchema(pyramid_request, "", "")

        appstruct = schema.validate({"userid": "new_userid"})

        assert "userid" not in appstruct
        assert "userid" not in appstruct.get("extra", {})

    def test_you_cannot_change_an_annotations_references(self, pyramid_request):
        schema = UpdateAnnotationSchema(pyramid_request, "", "")

        appstruct = schema.validate({"references": ["new_parent"]})

        assert "references" not in appstruct
        assert "references" not in appstruct.get("extra", {})

    def test_it_replaces_private_permissions_with_shared_False(self, pyramid_request):
        schema = UpdateAnnotationSchema(pyramid_request, "", "")

        appstruct = schema.validate(
            {"permissions": {"read": ["acct:harriet@example.com"]}}
        )

        assert appstruct["shared"] is False
        assert "permissions" not in appstruct
        assert "permissions" not in appstruct.get("extras", {})

    def test_it_replaces_shared_permissions_with_shared_True(self, pyramid_request):
        schema = UpdateAnnotationSchema(pyramid_request, "", "__world__")

        appstruct = schema.validate({"permissions": {"read": ["group:__world__"]}})

        assert appstruct["shared"] is True
        assert "permissions" not in appstruct
        assert "permissions" not in appstruct.get("extras", {})

    def test_it_passes_existing_target_uri_to_document_uris_from_data(
        self, document_claims, pyramid_request
    ):
        """
        If no 'uri' is given it should use the existing target_uri.

        If no 'uri' is given in the update request then
        document_uris_from_data() should be called with the existing
        target_uri of the annotation in the database.

        """
        document_data = {"foo": "bar"}
        schema = UpdateAnnotationSchema(pyramid_request, mock.sentinel.target_uri, "")

        schema.validate({"document": document_data})

        document_claims.document_uris_from_data.assert_called_once_with(
            document_data, claimant=mock.sentinel.target_uri
        )

    def test_it_passes_existing_target_uri_to_document_metas_from_data(
        self, document_claims, pyramid_request
    ):
        """
        If no 'uri' is given it should use the existing target_uri.

        If no 'uri' is given in the update request then
        document_metas_from_data() should be called with the existing
        target_uri of the annotation in the database.

        """
        document_data = {"foo": "bar"}
        schema = UpdateAnnotationSchema(pyramid_request, mock.sentinel.target_uri, "")

        schema.validate({"document": document_data})

        document_claims.document_metas_from_data.assert_called_once_with(
            document_data, claimant=mock.sentinel.target_uri
        )


@pytest.fixture
def document_claims(patch):
    return patch("hyputils.memex.schemas.annotation.document_claims")



# -*- coding: utf-8 -*-
"""
A module into which all ORM classes are imported.

To avoid circular imports almost all code should import ORM classes from this
module rather than importing them directly,
``from hyputils.memex import models`` rather than ``from hyputils.memex.foo import models``

This is a convenience - you can just import this one module and all of the
ORM classes will be defined, instead of having to import every models module
individually.

For example when testing ORM classes the test module for ``h.foo.models.Bar``
can't just import ``h.foo.models``, it would also need to import the models
module for each database table that ``Bar`` has a (direct or indirect) foreign
key to. So for convenience the test module can instead just do
``from hyputils.memex import models`` and have all ORM classes be defined.

"""
from __future__ import unicode_literals

from hyputils.memex.models.annotation import Annotation
from hyputils.memex.models.document import Document, DocumentMeta, DocumentURI
from hyputils.memex.models.group import Group
from hyputils.memex.models.organization import Organization
from hyputils.memex.models.user import User
from hyputils.memex.models.user_identity import UserIdentity

__all__ = (
    "Annotation",
    "Document",
    "DocumentMeta",
    "DocumentURI",
    "Group",
    "Organization",
    "User",
    "UserIdentity",
)



# -*- coding: utf-8 -*-

from __future__ import unicode_literals
from collections import namedtuple

import enum
import re
import sqlalchemy as sa
import slugify

from hyputils.memex.db import Base
from hyputils.memex.db import mixins
from hyputils.memex import pubid
from hyputils.memex.util.group import split_groupid
from hyputils.memex.security import security


GROUP_NAME_MIN_LENGTH = 3
GROUP_NAME_MAX_LENGTH = 25
GROUP_DESCRIPTION_MAX_LENGTH = 250
AUTHORITY_PROVIDED_ID_PATTERN = r"^[a-zA-Z0-9._\-+!~*()']+$"
AUTHORITY_PROVIDED_ID_MAX_LENGTH = 1024


class JoinableBy(enum.Enum):
    authority = "authority"


class ReadableBy(enum.Enum):
    members = "members"
    world = "world"


class WriteableBy(enum.Enum):
    authority = "authority"
    members = "members"


class Group(Base, mixins.Timestamps):
    __tablename__ = "group"

    __table_args__ = (
        # Add a composite index of the (authority, authority_provided_id)
        # columns. Also impose uniqueness such that no two records may share
        # the same (authority, authority_provided_id) composite
        #
        # See:
        #
        # * http://docs.sqlalchemy.org/en/latest/core/constraints.html#indexes
        sa.Index(
            "ix__group__groupid", "authority", "authority_provided_id", unique=True
        ),
    )

    id = sa.Column(sa.Integer, autoincrement=True, primary_key=True)
    # We don't expose the integer PK to the world, so we generate a short
    # random string to use as the publicly visible ID.
    pubid = sa.Column(sa.Text(), default=pubid.generate, unique=True, nullable=False)
    authority = sa.Column(sa.UnicodeText(), nullable=False)
    name = sa.Column(sa.UnicodeText(), nullable=False, index=True)

    creator_id = sa.Column(sa.Integer, sa.ForeignKey("user.id"))
    creator = sa.orm.relationship("User")

    description = sa.Column(sa.UnicodeText())

    #: Allow authorities to define their own unique identifier for a group
    #: (versus the pubid). This identifier is owned by the authority/client
    #: versus ``pubid``, which is owned and controlled by the service.
    authority_provided_id = sa.Column(sa.UnicodeText(), nullable=True)

    #: Which type of user is allowed to join this group, possible values are:
    #: authority, None
    joinable_by = sa.Column(
        sa.Enum(JoinableBy, name="group_joinable_by"), nullable=True
    )

    #: Which type of user is allowed to read annotations in this group,
    #: possible values are: authority, members, world
    readable_by = sa.Column(
        sa.Enum(ReadableBy, name="group_readable_by"), nullable=True, index=True
    )

    #: Which type of user is allowed to write to this group, possible values
    #: are: authority, members
    writeable_by = sa.Column(
        sa.Enum(WriteableBy, name="group_writeable_by"), nullable=True
    )

    @property
    def groupid(self):
        if self.authority_provided_id is None:
            return None
        return "group:{authority_provided_id}@{authority}".format(
            authority_provided_id=self.authority_provided_id, authority=self.authority
        )

    @groupid.setter
    def groupid(self, value):
        """
        Deconstruct a formatted ``groupid`` and set its constituent properties
        on the instance.

        If ``groupid`` is set to None, set ``authority_provided_id`` to None
        but leave authority untouched—this allows a caller to nullify the
        ``authority_provided_id`` field.

        :raises ValueError: if ``groupid`` is an invalid format
        """
        if value is None:
            self.authority_provided_id = None
        else:
            groupid_parts = split_groupid(value)
            self.authority_provided_id = groupid_parts["authority_provided_id"]
            self.authority = groupid_parts["authority"]

    # Group membership
    members = sa.orm.relationship(
        "User",
        secondary="user_group",
        backref=sa.orm.backref("groups", order_by="Group.name"),
    )

    organization_id = sa.Column(
        sa.Integer, sa.ForeignKey("organization.id"), nullable=True
    )
    organization = sa.orm.relationship("Organization")

    def __init__(self, **kwargs):
        super(Group, self).__init__(**kwargs)

    @sa.orm.validates("name")
    def validate_name(self, key, name):
        if not GROUP_NAME_MIN_LENGTH <= len(name) <= GROUP_NAME_MAX_LENGTH:
            raise ValueError(
                "name must be between {min} and {max} characters "
                "long".format(min=GROUP_NAME_MIN_LENGTH, max=GROUP_NAME_MAX_LENGTH)
            )
        return name

    @sa.orm.validates("authority_provided_id")
    def validate_authority_provided_id(self, key, authority_provided_id):
        if not authority_provided_id:
            return None

        if not re.match(AUTHORITY_PROVIDED_ID_PATTERN, authority_provided_id):
            raise ValueError(
                "authority_provided_id must only contain characters allowed"
                r" in encoded URIs: [a-zA-Z0-9._\-+!~*()']"
            )

        if len(authority_provided_id) > AUTHORITY_PROVIDED_ID_MAX_LENGTH:
            raise ValueError(
                "authority_provided_id must be {max} characters or fewer"
                "characters long".format(max=AUTHORITY_PROVIDED_ID_MAX_LENGTH)
            )

        return authority_provided_id

    @property
    def slug(self):
        """A version of this group's name suitable for use in a URL."""
        return slugify.slugify(self.name)

    @property
    def type(self):
        """
        The "type" of this group, e.g. "open" or "private".

        :rtype: string
        :raises ValueError: if the type of the group isn't recognized

        """
        self_type_flags = TypeFlags(
            joinable_by=self.joinable_by,
            readable_by=self.readable_by,
            writeable_by=self.writeable_by,
        )

        for type_, type_flags in (
            ("open", OPEN_GROUP_TYPE_FLAGS),
            ("private", PRIVATE_GROUP_TYPE_FLAGS),
            ("restricted", RESTRICTED_GROUP_TYPE_FLAGS),
        ):
            if self_type_flags == type_flags:
                return type_

        raise ValueError(
            "This group doesn't seem to match any known type of group. "
            "This shouldn't be in the database!"
        )

    @property
    def is_public(self):
        return self.readable_by == ReadableBy.world

    def __acl__(self):
        terms = []

        join_principal = _join_principal(self)
        if join_principal is not None:
            terms.append((security.Allow, join_principal, "join"))

        read_principal = _read_principal(self)
        if read_principal is not None:
            terms.append((security.Allow, read_principal, "read"))

        flag_principal = _flag_principal(self)
        if flag_principal is not None:
            terms.append((security.Allow, flag_principal, "flag"))

        write_principal = _write_principal(self)
        if write_principal is not None:
            terms.append((security.Allow, write_principal, "write"))

        if self.creator:
            # The creator of the group should be able to update it
            terms.append((security.Allow, self.creator.userid, "admin"))
            terms.append((security.Allow, self.creator.userid, "moderate"))
            # The creator may update this group in an upsert context
            terms.append((security.Allow, self.creator.userid, "upsert"))

        # This authority principal may be used to grant auth clients
        # permissions for groups within their authority
        authority_principal = "client_authority:{}".format(self.authority)

        # auth_clients that have the same authority as the target group
        # may add members to it
        terms.append((security.Allow, authority_principal, "member_add"))
        # auth_clients that have the same authority as this group
        # should be allowed to update it
        terms.append((security.Allow, authority_principal, "admin"))
        # auth_clients with matching authority should be able to read
        # the group
        terms.append((security.Allow, authority_principal, "read"))

        terms.append(security.DENY_ALL)

        return terms

    def __repr__(self):
        return "<Group: %s>" % self.slug

    @classmethod
    def created_by(cls, session, user):
        """Return a query object filtering groups by creator."""
        return session.query(cls).filter(Group.creator == user)


def _join_principal(group):
    return {JoinableBy.authority: "authority:{}".format(group.authority)}.get(
        group.joinable_by
    )


def _read_principal(group):
    return {
        ReadableBy.members: "group:{}".format(group.pubid),
        ReadableBy.world: security.Everyone,
    }.get(group.readable_by)


def _flag_principal(group):
    # If a user can read (see) annotations within this group,
    # they can also flag them—but they need to be logged in
    # (``pyramid.security.Authenticated``)
    return {
        ReadableBy.members: "group:{}".format(group.pubid),
        ReadableBy.world: security.Authenticated,
    }.get(group.readable_by)


def _write_principal(group):
    return {
        WriteableBy.authority: "authority:{}".format(group.authority),
        WriteableBy.members: "group:{}".format(group.pubid),
    }.get(group.writeable_by)


TypeFlags = namedtuple("TypeFlags", "joinable_by readable_by writeable_by")


OPEN_GROUP_TYPE_FLAGS = TypeFlags(
    joinable_by=None, readable_by=ReadableBy.world, writeable_by=WriteableBy.authority
)


PRIVATE_GROUP_TYPE_FLAGS = TypeFlags(
    joinable_by=JoinableBy.authority,
    readable_by=ReadableBy.members,
    writeable_by=WriteableBy.members,
)


RESTRICTED_GROUP_TYPE_FLAGS = TypeFlags(
    joinable_by=None, readable_by=ReadableBy.world, writeable_by=WriteableBy.members
)


USER_GROUP_TABLE = sa.Table(
    "user_group",
    Base.metadata,
    sa.Column("id", sa.Integer, autoincrement=True, primary_key=True),
    sa.Column("user_id", sa.Integer, sa.ForeignKey("user.id"), nullable=False),
    sa.Column("group_id", sa.Integer, sa.ForeignKey("group.id"), nullable=False),
    sa.UniqueConstraint("user_id", "group_id"),
)



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import factory

from hyputils.memex import models

from .base import FAKER, ModelFactory


class Document(ModelFactory):
    class Meta:
        model = models.Document


class DocumentMeta(ModelFactory):
    class Meta:
        model = models.DocumentMeta

    # Trying to add two DocumentMetas with the same claimant and type to the
    # db will crash. We use a sequence instead of something like FAKER.url()
    # for claimant here so that never happens (unless you pass in your own
    # claimant).
    claimant = factory.Sequence(lambda n: "http://example.com/document_" + str(n) + "/")

    type = factory.Iterator(
        ["title", "twitter.url.main_url", "twitter.title", "favicon"]
    )
    document = factory.SubFactory(Document)

    @factory.lazy_attribute
    def value(self):
        if self.type == "twitter.url.main_url":
            return [FAKER.url()]
        elif self.type == "favicon":
            return [FAKER.image_url()]
        else:
            return [FAKER.bs()]


class DocumentURI(ModelFactory):
    class Meta:
        model = models.DocumentURI

    # Trying to add two DocumentURIs with the same claimant, uri, type and
    # content_type to the db will crash. We use a sequence instead of something
    # like FAKER.url() for claimant here so that never happens (unless you pass
    # in your own claimant).
    claimant = factory.Sequence(lambda n: "http://example.com/document_" + str(n) + "/")

    uri = factory.LazyAttribute(lambda obj: obj.claimant)
    type = factory.Iterator(
        ["rel-alternate", "rel-canonical", "highwire-pdf", "dc-doi"]
    )
    content_type = factory.Iterator(["text/html", "application/pdf", "text/plain"])
    document = factory.SubFactory(Document)



# -*- coding: utf-8 -*-

"""Classes for validating data supplied to the application."""
from __future__ import unicode_literals

from hyputils.memex.schemas.base import ValidationError

__all__ = ("ValidationError",)



# -*- coding: utf-8 -*-

from __future__ import unicode_literals
import sqlalchemy as sa

from hyputils.memex.db import Base
from hyputils.memex.db import mixins
from hyputils.memex import pubid

ORGANIZATION_DEFAULT_PUBID = "__default__"
ORGANIZATION_NAME_MIN_CHARS = 1
ORGANIZATION_NAME_MAX_CHARS = 25


class Organization(Base, mixins.Timestamps):
    __tablename__ = "organization"

    id = sa.Column(sa.Integer, autoincrement=True, primary_key=True)

    # We don't expose the integer PK to the world, so we generate a short
    # random string to use as the publicly visible ID.
    pubid = sa.Column(sa.Text(), default=pubid.generate, unique=True, nullable=False)

    name = sa.Column(sa.UnicodeText(), nullable=False, index=True)

    logo = sa.Column(sa.UnicodeText())

    authority = sa.Column(sa.UnicodeText(), nullable=False)

    @sa.orm.validates("name")
    def validate_name(self, key, name):
        if not (
            ORGANIZATION_NAME_MIN_CHARS <= len(name) <= ORGANIZATION_NAME_MAX_CHARS
        ):
            raise ValueError(
                "name must be between {min} and {max} characters long".format(
                    min=ORGANIZATION_NAME_MIN_CHARS, max=ORGANIZATION_NAME_MAX_CHARS
                )
            )
        return name

    def __repr__(self):
        return "<Organization: %s>" % self.pubid

    @classmethod
    def default(cls, session):
        return session.query(cls).filter_by(pubid=ORGANIZATION_DEFAULT_PUBID).one()



# -*- coding: utf-8 -*-

from __future__ import unicode_literals
import pytest

from hyputils.memex.util import user as user_util


def test_split_user():
    parts = user_util.split_user("acct:seanh@hypothes.is")
    assert parts == {"username": "seanh", "domain": "hypothes.is"}


def test_split_user_no_match():
    with pytest.raises(ValueError):
        user_util.split_user("donkeys")



def database_url(url):
    """Parse a string as a Heroku-style database URL."""
    # Heroku database URLs start with postgres://, which is an old and
    # deprecated dialect as far as sqlalchemy is concerned. We upgrade this
    # to postgresql+psycopg2 by default.
    if url.startswith("postgres://"):
        url = "postgresql+psycopg2://" + url[len("postgres://") :]
    return url



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import factory

from hyputils.memex import models

from .base import ModelFactory


class UserIdentity(ModelFactory):
    class Meta:
        model = models.UserIdentity
        sqlalchemy_session_persistence = "flush"

    provider = factory.Sequence(lambda n: "test_provider_{n}".format(n=str(n)))
    provider_unique_id = factory.Sequence(lambda n: "test_id_{n}".format(n=str(n)))



# -*- encoding: utf-8 -*-

"""
Functions for parsing document claims data from the client.

Functions for parsing the document claims (document metadata claims and URI
equivalence claims) that the client POSTS in the JSON "document" sub-object in
annotation create and update requests.

The data is parsed into a format suitable for storage in our database model,
and returned.

"""

from __future__ import unicode_literals

import re


# Pattern that matches DOIs (eg. "10.1000/123456").
#
# See https://www.doi.org/doi_handbook/2_Numbering.html#2.2 The registrant code
# is not specified as numeric there but is currently always a 4+ digit value.
# See https://www.doi.org/overview/DOI_article_ELIS3.pdf.
#
# This also accepts DOIs represented as URLs (eg.
# "https://doi.org/10.1000/123456").
DOI_PATTERN = re.compile(r"(https?://(dx\.)?doi\.org/)?10\.[0-9]{4,}[.0-9]*/.*")


def document_uris_from_data(document_data, claimant):
    """
    Return one or more document URI dicts for the given document data.

    Returns one document uri dict for each document equivalence claim in
    document_data.

    Each dict can be used to init a DocumentURI object directly::

        document_uri = DocumentURI(**document_uri_dict)

    Always returns at least one "self-claim" document URI whose URI is the
    claimant URI itself.

    :param document_data: the "document" sub-object that was POSTed to the API
        as part of a new or updated annotation
    :type document_data: dict

    :param claimant: the URI that the browser was at when this annotation was
        created (the top-level "uri" field of the annotation)
    :type claimant: unicode

    :returns: a list of one or more document URI dicts
    :rtype: list of dicts

    """
    document_uris = document_uris_from_links(document_data.get("link", []), claimant)

    document_uris.extend(
        document_uris_from_highwire_pdf(document_data.get("highwire", {}), claimant)
    )

    document_uris.extend(
        document_uris_from_highwire_doi(document_data.get("highwire", {}), claimant)
    )

    document_uris.extend(document_uris_from_dc(document_data.get("dc", {}), claimant))

    document_uris.append(document_uri_self_claim(claimant))

    for document_uri in document_uris:
        uri = document_uri["uri"]
        if uri:
            document_uri["uri"] = uri.strip()

    document_uris = [d for d in document_uris if d["uri"]]

    return document_uris


def document_metas_from_data(document_data, claimant):
    """
    Return a list of document meta dicts for the given document data.

    Returns one document meta dict for each document metadata claim in
    document_data.

    Each dict can be used to init a DocumentMeta object directly::

        document_meta = DocumentMeta(**document_meta_dict)

    :param document_data: the "document" sub-object that the client POSTed to
        the API as part of a new or updated annotation
    :type document_data: dict

    :param claimant: the URI that the browser was at when this annotation was
        created (the top-level "uri" field of the annotation)
    :type claimant: unicode

    :returns: a list of zero or more document meta dicts
    :rtype: list of dicts

    """

    def transform_meta_(document_meta_dicts, items, path_prefix=None):
        """Fill document_meta_dicts with document meta dicts for the items."""
        if path_prefix is None:
            path_prefix = []

        for key, value in items.items():
            keypath = path_prefix[:]
            keypath.append(key)

            if isinstance(value, dict):
                transform_meta_(document_meta_dicts, value, path_prefix=keypath)
            else:
                if not isinstance(value, list):
                    value = [value]

                type_ = ".".join(keypath)

                if type_ == "title":
                    # We don't allow None, empty strings, whitespace-only
                    # strings, leading or trailing whitespaces, or empty arrays
                    # in document title values.
                    value = [v.strip() for v in value if v and v.strip()]
                    if not value:
                        continue

                document_meta_dicts.append(
                    {"type": type_, "value": value, "claimant": claimant}
                )

    items = {k: v for k, v in document_data.items() if k != "link"}
    document_meta_dicts = []
    transform_meta_(document_meta_dicts, items)
    return document_meta_dicts


def document_uris_from_links(link_dicts, claimant):
    """
    Return document URI dicts for the given document.link data.

    Process a document.link list of dicts that the client submitted as part of
    an annotation create or update request and return document URI dicts for
    all of the document equivalence claims that it makes.

    """
    document_uris = []
    for link in link_dicts:

        # In Py 3, `dict.keys()` does not return a list and cannot be compared
        # with a list, so explicitly convert the result.
        link_keys = list(link.keys())

        # Disregard self-claim URLs as they're added separately later.
        if link_keys == ["href"] and link["href"] == claimant:
            continue

        # Disregard doi links as these are being added separately from the
        # highwire and dc metadata later on.
        if link_keys == ["href"] and link["href"].startswith("doi:"):
            continue

        # Disregard Highwire PDF links as these are being added separately from
        # the highwire metadata later on.
        if set(link_keys) == set(["href", "type"]):
            if link["type"] == "application/pdf":
                continue

        uri_ = link["href"]

        # Handle rel="..." links.
        if "rel" in link:
            type_ = "rel-{}".format(link["rel"])
        else:
            type_ = ""

        # The "type" item in link dicts becomes content_type in DocumentURIs.
        content_type = link.get("type", "")

        document_uris.append(
            {
                "claimant": claimant,
                "uri": uri_,
                "type": type_,
                "content_type": content_type,
            }
        )

    return document_uris


def document_uris_from_highwire_pdf(highwire_dict, claimant):
    """
    Return PDF document URI dicts for the given 'highwire' document metadata.

    Process a document.highwire dict that the client submitted as part of an
    annotation create or update request and return document URI dicts for all
    of the PDF document equivalence claims that it makes.

    """
    document_uris = []
    hwpdfvalues = highwire_dict.get("pdf_url", [])
    for pdf in hwpdfvalues:
        document_uris.append(
            {
                "claimant": claimant,
                "uri": pdf,
                "type": "highwire-pdf",
                "content_type": "application/pdf",
            }
        )
    return document_uris


def document_uris_from_highwire_doi(highwire_dict, claimant):
    """
    Return DOI document URI dicts for the given 'highwire' document metadata.

    Process a document.highwire dict that the client submitted as part of an
    annotation create or update request and return document URI dicts for all
    of the 'doi:' document equivalence claims that it makes.

    """
    document_uris = []
    hwdoivalues = highwire_dict.get("doi", [])
    for doi in hwdoivalues:
        doi = doi_uri_from_string(doi)
        if doi is not None:
            document_uris.append(
                {
                    "claimant": claimant,
                    "uri": doi,
                    "type": "highwire-doi",
                    "content_type": "",
                }
            )
    return document_uris


def document_uris_from_dc(dc_dict, claimant):
    """
    Return document URI dicts for the given 'dc' document metadata.

    Process a document.dc dict that the client submitted as part of an
    annotation create or update request and return document URI dicts for all
    of the document equivalence claims that it makes.

    """
    document_uris = []
    dcdoivalues = dc_dict.get("identifier", [])
    for doi in dcdoivalues:
        doi = doi_uri_from_string(doi)
        if doi is not None:
            document_uris.append(
                {"claimant": claimant, "uri": doi, "type": "dc-doi", "content_type": ""}
            )

    return document_uris


def document_uri_self_claim(claimant):
    """Return a "self-claim" document URI dict for the given claimant."""
    return {
        "claimant": claimant,
        "uri": claimant,
        "type": "self-claim",
        "content_type": "",
    }


def doi_uri_from_string(s):
    """
    Return the DOI URI from the given user-supplied string, or None.

    Return a string of the format "doi:<id>". Leading and trailing whitespace
    is stripped from the string as a whole and from the <id> substring.

    If the given string doesn't already start with "doi:" it is prepended.

    If the given string, minus any `doi:` prefix does not match the syntax
    for DOI names ("10.NNNN/.*") then `None` is returned.
    """
    s = s.strip()

    if s.startswith("doi:"):
        s = s[len("doi:") :]

    s = s.strip()

    if DOI_PATTERN.match(s) is None:
        return None

    s = "doi:{}".format(s)

    return s



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import factory

from hyputils.memex import models

from .base import ModelFactory


class Organization(ModelFactory):
    class Meta:
        model = models.Organization
        sqlalchemy_session_persistence = "flush"

    name = factory.Sequence(lambda n: "Test Organization {n}".format(n=str(n)))
    authority = "example.com"



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import re
from functools import partial

import bleach
import mistune
from bleach.linkifier import LinkifyFilter

LINK_REL = "nofollow noopener"

MARKDOWN_TAGS = [
    "a",
    "blockquote",
    "code",
    "em",
    "hr",
    "h1",
    "h2",
    "h3",
    "h4",
    "h5",
    "h6",
    "img",
    "li",
    "ol",
    "p",
    "pre",
    "strong",
    "ul",
]
ALLOWED_TAGS = set(bleach.ALLOWED_TAGS + MARKDOWN_TAGS)


def _filter_link_attributes(tag, name, value):
    if name in ["href", "title"]:
        return True

    if name == "target" and value == "_blank":
        return True

    if name == "rel" and value == LINK_REL:
        return True

    return False


MARKDOWN_ATTRIBUTES = {"a": _filter_link_attributes, "img": ["alt", "src", "title"]}

ALLOWED_ATTRIBUTES = bleach.ALLOWED_ATTRIBUTES.copy()
ALLOWED_ATTRIBUTES.update(MARKDOWN_ATTRIBUTES)

# Singleton instance of the bleach cleaner
cleaner = None
# Singleton instance of the Markdown instance
markdown = None


class MathMarkdown(mistune.Markdown):
    def output_block_math(self):
        return self.renderer.block_math(self.token["text"])


class MathInlineLexer(mistune.InlineLexer):
    def __init__(self, *args, **kwargs):
        super(MathInlineLexer, self).__init__(*args, **kwargs)
        self.rules.inline_math = re.compile(r"\\\((.*?)\\\)", re.DOTALL)
        self.default_rules.insert(0, "inline_math")

    def output_inline_math(self, m):
        return self.renderer.inline_math(m.group(1))


class MathBlockLexer(mistune.BlockLexer):
    def __init__(self, *args, **kwargs):
        super(MathBlockLexer, self).__init__(*args, **kwargs)
        self.rules.block_math = re.compile(r"^\$\$(.*?)\$\$", re.DOTALL)
        self.default_rules.insert(0, "block_math")

    def parse_block_math(self, m):
        self.tokens.append({"type": "block_math", "text": m.group(1)})


class MathRenderer(mistune.Renderer):
    def __init__(self, **kwargs):
        super(MathRenderer, self).__init__(**kwargs)

    def block_math(self, text):
        return "<p>$$%s$$</p>\n" % text

    def inline_math(self, text):
        return "\\(%s\\)" % text


def render(text):
    if text is not None:
        render = _get_markdown()
        return sanitize(render(text))
    return None


def sanitize(text):
    cleaner = _get_cleaner()
    return cleaner.clean(text)


def _linkify_target_blank(attrs, new=False):
    # FIXME: when bleach>2.0.0 is released we can use
    # bleach.callbacks.target_blank instead of this function. We have our own
    # copy to work around a bug in 2.0.0:
    #
    #   https://github.com/mozilla/bleach/commit/b23c74c1ca5ffcbd308df93e79487fa92a6eb4a7
    #
    href_key = (None, "href")

    if href_key not in attrs:
        return attrs

    if attrs[href_key].startswith("mailto:"):
        return attrs

    attrs[(None, "target")] = "_blank"
    return attrs


def _linkify_rel(attrs, new=False):
    href_key = (None, "href")

    if href_key not in attrs:
        return attrs

    if attrs[href_key].startswith("mailto:"):
        return attrs

    attrs[(None, "rel")] = LINK_REL
    return attrs


def _get_cleaner():
    global cleaner
    if cleaner is None:
        linkify_filter = partial(
            LinkifyFilter, callbacks=[_linkify_target_blank, _linkify_rel]
        )
        cleaner = bleach.Cleaner(
            tags=ALLOWED_TAGS, attributes=ALLOWED_ATTRIBUTES, filters=[linkify_filter]
        )
    return cleaner


def _get_markdown():
    global markdown
    if markdown is None:
        markdown = MathMarkdown(
            renderer=MathRenderer(),
            inline=MathInlineLexer,
            block=MathBlockLexer,
            escape=True,
        )
    return markdown



import os
import unittest
from time import sleep
import pytest
from hyputils.handlers import annotationSyncHandler
from hyputils.subscribe import AnnotationStream, preFilter
from hyputils.hypothesis import group

# TODO use our own little websocket server to send annos as if they came from hypothesis

class FakeMem:
    def memoize_annos(self, annos):
        print([anno._row for anno in annos])


class TestStream(unittest.TestCase):
    def test_0_stream(self):
        annos = []
        prefilter = preFilter(groups=['__world__', group], users=['tgbugstest']).export()
        annoStream = AnnotationStream(annos,
                                      prefilter,
                                      annotationSyncHandler)
        stream_thread, exit_loop = annoStream()
        stream_thread.start()
        sleep(1)
        exit_loop()
        stream_thread.join()

    @pytest.mark.skip(reason='local test interactive and blocks')
    def test_9999_get_from_ws(self):
        from IPython import embed
        fake_mem = FakeMem()
        annos = []
        prefilter = preFilter(groups=['__world__', group]).export()
        annoStream = AnnotationStream(annos,
                                      prefilter,
                                      annotationSyncHandler,
                                      memoizer=fake_mem)
        stream_thread, exit_loop = annoStream()
        stream_thread.start()
        embed()
        exit_loop()
        stream_thread.join()



#!/usr/bin/env python3.6
"""
Usage:
    test <name>
"""

# WS client example

import asyncio
import websockets
from docopt import docopt
args = docopt(__doc__)
name = args['<name>']

async def hello():
    async with websockets.connect('ws://localhost:5050') as websocket:
        await websocket.send(name)
        print(f"> {name}")

        greeting = await websocket.recv()
        print(f"< {greeting}")
        while True:
            print('waiting for next message')
            print(await websocket.recv())

asyncio.get_event_loop().run_until_complete(hello())





# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import datetime
import random
import uuid

import factory
from sqlalchemy import orm

from hyputils.memex import models
from hyputils.memex.models.document import update_document_metadata

from .base import FAKER, ModelFactory
from .document import Document, DocumentMeta, DocumentURI


class Annotation(ModelFactory):
    class Meta:
        model = models.Annotation
        sqlalchemy_session_persistence = (
            "flush"
        )  # Always flush the db to generate annotation.id.

    tags = factory.LazyFunction(lambda: FAKER.words(nb=random.randint(0, 5)))
    target_uri = factory.Faker("uri")
    text = factory.Faker("paragraph")
    userid = factory.LazyFunction(
        lambda: "acct:{username}@{authority}".format(
            username=FAKER.user_name(), authority=FAKER.domain_name(levels=1)
        )
    )
    document = factory.SubFactory(Document)
    groupid = "__world__"

    @factory.lazy_attribute
    def target_selectors(self):
        return [
            {
                "endContainer": "/div[1]/article[1]/section[1]/div[1]/div[2]/div[1]",
                "endOffset": 76,
                "startContainer": "/div[1]/article[1]/section[1]/div[1]/div[2]/div[1]",
                "startOffset": 0,
                "type": "RangeSelector",
            },
            {"end": 362, "start": 286, "type": "TextPositionSelector"},
            {
                "exact": "If you wish to install Hypothesis on your own site then head over to GitHub.",
                "prefix": " browser extension.\n            ",
                "suffix": "\n          \n        \n      \n    ",
                "type": "TextQuoteSelector",
            },
        ]

    @factory.post_generation
    def make_metadata(self, create, extracted, **kwargs):
        """Create associated document metadata for the annotation."""
        # The metadata objects are going to be added to the db, so if we're not
        # using the create strategy then simply don't make any.
        if not create:
            return

        def document_uri_dict():
            """
            Return a randomly generated DocumentURI dict for this annotation.

            This doesn't add anything to the database session yet.
            """
            document_uri = DocumentURI.build(
                document=None, claimant=self.target_uri, uri=self.target_uri
            )
            return dict(
                claimant=document_uri.claimant,
                uri=document_uri.uri,
                type=document_uri.type,
                content_type=document_uri.content_type,
            )

        document_uri_dicts = [document_uri_dict() for _ in range(random.randint(1, 3))]

        def document_meta_dict(type_=None):
            """
            Return a randomly generated DocumentMeta dict for this annotation.

            This doesn't add anything to the database session yet.
            """
            kwargs = {"document": None, "claimant": self.target_uri}

            if type_ is not None:
                kwargs["type"] = type_

            document_meta = DocumentMeta.build(**kwargs)

            return dict(
                claimant=document_meta.claimant,
                type=document_meta.type,
                value=document_meta.value,
            )

        document_meta_dicts = [
            document_meta_dict() for _ in range(random.randint(1, 3))
        ]

        # Make sure that there's always at least one DocumentMeta with
        # type='title', so that we never get annotation.document.title is None:
        if "title" not in [m["type"] for m in document_meta_dicts]:
            document_meta_dicts.append(document_meta_dict(type_="title"))

        self.document = update_document_metadata(
            orm.object_session(self),
            self.target_uri,
            document_meta_dicts=document_meta_dicts,
            document_uri_dicts=document_uri_dicts,
            created=self.created,
            updated=self.updated,
        )

    @factory.post_generation
    def make_id(self, create, extracted, **kwargs):
        """Add a randomly ID if the annotation doesn't have one yet."""
        # If using the create strategy don't generate an id.
        # models.Annotation.id's server_default function will generate one
        # when the annotation is saved to the DB.
        if create:
            return

        # Don't generate an id if the user passed in one of their own.
        if getattr(self, "id", None):
            return

        self.id = uuid.uuid4().hex

    @factory.post_generation
    def timestamps(self, create, extracted, **kwargs):
        # If using the create strategy let sqlalchemy set the created and
        # updated times when saving to the DB.
        if create:
            return

        # When using the build or stub strategy sqlalchemy won't set created or updated
        # times for us, so do it ourselves instead.
        #
        # We're generating created and updated separately (calling now() twice
        # instead of just once) so created and updated won't be exactly the
        # same. This is consistent with how models.Annotation does it when
        # saving to the DB.
        self.created = self.created or datetime.datetime.now()
        self.updated = self.updated or datetime.datetime.now()



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import pytest
from hyputils.memex import models


def test_init_sets_given_attributes():
    logo = """<svg width="100" height="100" xmlns="http://www.w3.org/2000/svg">
            <circle cx="50" cy="50" r="40" stroke="green" stroke-width="4" fill="yellow" />
            </svg>"""
    organization = models.Organization(
        name="My organization", logo=logo, authority="example.com"
    )

    assert organization.name == "My organization"
    assert organization.logo == logo
    assert organization.authority == "example.com"


def test_null_logo():
    organization = models.Organization(name="My Organization")

    assert organization.name == "My Organization"
    assert organization.logo is None


def test_too_short_name_raises_value_error():
    with pytest.raises(ValueError):
        models.Organization(name="")


def test_too_long_name_raises_value_error():
    with pytest.raises(ValueError):
        models.Organization(name="abcdefghijklmnopqrstuvwxyz")


def test_none_logo_is_valid():
    org = models.Organization(name="My Organization", logo=None)
    assert org.logo is None


def test_repr(db_session, factories):
    organization = models.Organization(
        name="My Organization", authority="example.com", pubid="test_pubid"
    )

    assert repr(organization) == "<Organization: test_pubid>"


def test_default_returns_the_default_organization(db_session):
    assert models.Organization.default(db_session).pubid == "__default__"



# -*- coding: utf-8 -*-
"""Some shared utility functions for manipulating group data."""
from __future__ import unicode_literals

import re

GROUPID_PATTERN = r"^group:([a-zA-Z0-9._\-+!~*()']{1,1024})@(.*)$"


def split_groupid(groupid):
    """Return the ``authority_provided_id`` and ``authority`` from a ``groupid``

    For example if groupid is u'group:339ae9f33c@myauth.org' then return
    {'authority_provided_id': u'339ae9f33c', 'authority': u'myauth.org'}'

    :raises ValueError: if the given groupid isn't a valid groupid

    """
    match = re.match(GROUPID_PATTERN, groupid)
    if match:
        return {
            "authority_provided_id": match.groups()[0],
            "authority": match.groups()[1],
        }
    raise ValueError("{groupid} isn't a valid groupid".format(groupid=groupid))


def is_groupid(maybe_groupid):
    """Return True if the given string looks like a groupid, else False."""
    return re.match(GROUPID_PATTERN, maybe_groupid) is not None
